import numpy as np
import numpy.fft as fft
import scipy.linalg as la
import scipy.ndimage as ndimage
import scipy.interpolate as sinterp
from scipy.stats import t
import warnings

def make_polar_coordinates(x, y, center=[0,0]):
    '''
    Args:
        x: meshgrid of x coordinates
        y: meshgrid of y coordinates
        center: new location of origin

    Returns:
        polar coordinates centered at the specified origin
    '''

    r = np.sqrt((x - center[0]) ** 2 + (y - center[1]) ** 2)
    phi = np.arctan2(y - center[1], x - center[0])
    # make sure phi is in range [-pi, pi)
    phi = (phi % (2 * np.pi)) - np.pi

    return r, phi


def collapse_data(data, pixel_weights=None, axis=1, collapse_method='mean'):
    """
    Function to collapse multi-dimensional data along axis using collapse_method

    Args:
        data: (multi-dimension)arrays of 2D images or 3D cubes.
        pixel_weights: ones if collapse method is not weighted collapse
        axis: axis index along which to collapse
        collapse_method: currently support 'median', 'mean', 'weighted-mean', 'trimmed-mean', 'weighted-median'

    Returns:
        Collapsed data
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)

        if collapse_method.lower() == 'median':

            return np.nanmedian(data, axis=axis)

        elif collapse_method.lower() == 'mean':

            return np.nanmean(data, axis=axis)



        elif ('weighted' in collapse_method.lower()) and ('mean' in collapse_method.lower()):
            if pixel_weights is None:
                pixel_weights = np.ones(data.shape)
            collapsed_data = np.nanmean(pixel_weights * data, axis=axis)
            collapsed_data /= np.nanmean(pixel_weights, axis=axis)

            return collapsed_data

        elif ('trimmed' in collapse_method.lower()) and ('mean' in collapse_method.lower()):
            collapsed = np.sort(data, axis=axis)
            collapsed = np.take(collapsed, range(2, collapsed.shape[axis] - 2), axis=axis)
            collapsed = np.nanmean(collapsed, axis=axis)

            return collapsed

        elif ('weighted' in collapse_method.lower()) and ('median' in collapse_method.lower()):
            if pixel_weights is None:
                pixel_weights = np.ones(data.shape)
            collapsed_data = np.nanmedian(pixel_weights * data, axis=axis)
            collapsed_data /= np.nanmedian(pixel_weights, axis=axis)

            return collapsed_data

        else:
            # default to mean collapse if input does not match any supported pattern
            print('{} collapse not supported, using mean collapse instead...'.format(collapse_method))

            return np.nanmean(data, axis=axis)


def klip_math(sci, ref_psfs, numbasis, covar_psfs=None, return_basis=False, return_basis_and_eig=False):
    """
    Helper function for KLIP that does the linear algebra
    
    Args:
        sci: array of length p containing the science data
        ref_psfs: N x p array of the N reference PSFs that 
                  characterizes the PSF of the p pixels
        numbasis: number of KLIP basis vectors to use (can be an int or an array of ints of length b)
        covar_psfs: covariance matrix of reference psfs passed in so you don't have to calculate it here
        return_basis: If true, return KL basis vectors (used when onesegment==True)
        return_basis_and_eig: If true, return KL basis vectors as well as the eigenvalues and eigenvectors of the
                                covariance matrix. Used for KLIP Forward Modelling of Laurent Pueyo.

    Returns:
        sub_img_rows_selected: array of shape (p,b) that is the PSF subtracted data for each of the b KLIP basis
                               cutoffs. If numbasis was an int, then sub_img_row_selected is just an array of length p
        KL_basis: array of shape (max(numbasis),p). Only if return_basis or return_basis_and_eig is True.
        evals: Eigenvalues of the covariance matrix. The covariance matrix is assumed NOT to be normalized by (p-1).
                Only if return_basis_and_eig is True.
        evecs: Eigenvectors of the covariance matrix. The covariance matrix is assumed NOT to be normalized by (p-1).
                Only if return_basis_and_eig is True.
    """
    # for the science image, subtract the mean and mask bad pixels
    sci_mean_sub = sci - np.nanmean(sci)
    # sci_nanpix = np.where(np.isnan(sci_mean_sub))
    # sci_mean_sub[sci_nanpix] = 0

    # do the same for the reference PSFs
    # playing some tricks to vectorize the subtraction
    ref_psfs_mean_sub = ref_psfs - np.nanmean(ref_psfs, axis=1)[:, None]
    ref_psfs_mean_sub[np.where(np.isnan(ref_psfs_mean_sub))] = 0

    # calculate the covariance matrix for the reference PSFs
    # note that numpy.cov normalizes by p-1 to get the NxN covariance matrix
    # we have to correct for that a few lines down when consturcting the KL
    # vectors since that's not part of the equation in the KLIP paper
    if covar_psfs is None:
        covar_psfs = np.cov(ref_psfs_mean_sub)

    # maximum number of KL modes
    tot_basis = covar_psfs.shape[0]

    # only pick numbasis requested that are valid. We can't compute more KL basis than there are reference PSFs
    # do numbasis - 1 for ease of indexing since index 0 is using 1 KL basis vector
    numbasis = np.clip(numbasis - 1, 0, tot_basis-1)  # clip values, for output consistency we'll keep duplicates
    max_basis = np.max(numbasis) + 1  # maximum number of eigenvectors/KL basis we actually need to use/calculate

    # calculate eigenvalues and eigenvectors of covariance matrix, but only the ones we need (up to max basis)
    evals, evecs = la.eigh(covar_psfs, eigvals=(tot_basis-max_basis, tot_basis-1))

    # check if there are negative eignevalues as they will cause NaNs later that we have to remove
    # the eigenvalues are ordered smallest to largest
    #check_nans = evals[-1] < 0 # currently this checks that *all* the evals are neg, but we want just one.
    # also, include 0 because that is a bad value too
    check_nans = np.any(evals <= 0) # alternatively, check_nans = evals[0] <= 0

    # scipy.linalg.eigh spits out the eigenvalues/vectors smallest first so we need to reverse
    # we're going to recopy them to hopefully improve caching when doing matrix multiplication
    evals = np.copy(evals[::-1])
    evecs = np.copy(evecs[:,::-1], order='F') #fortran order to improve memory caching in matrix multiplication

    # keep an index of the negative eignevalues for future reference if there are any
    if check_nans:
        neg_evals = (np.where(evals <= 0))[0]

    # calculate the KL basis vectors
    kl_basis = np.dot(ref_psfs_mean_sub.T, evecs)
    # JB question: Why is there this [None, :]? (It adds an empty first dimension)
    kl_basis = kl_basis * (1. / np.sqrt(evals * (np.size(sci) - 1)))[None, :]  #multiply a value for each row

    # sort to KL basis in descending order (largest first)
    # kl_basis = kl_basis[:,eig_args_all]

    # duplicate science image by the max_basis to do simultaneous calculation for different k_KLIP
    sci_mean_sub_rows = np.tile(sci_mean_sub, (max_basis, 1))
    sci_rows_selected = np.tile(sci_mean_sub, (np.size(numbasis), 1)) # this is the output image which has less rows

    # bad pixel mask
    # do it first for the image we're just doing computations on but don't care about the output
    sci_nanpix = np.where(np.isnan(sci_mean_sub_rows))
    sci_mean_sub_rows[sci_nanpix] = 0
    # now do it for the output image
    sci_nanpix = np.where(np.isnan(sci_rows_selected))
    sci_rows_selected[sci_nanpix] = 0

    # do the KLIP equation, but now all the different k_KLIP simultaneously
    # calculate the inner product of science image with each of the different kl_basis vectors
    # TODO: can we optimize this so it doesn't have to multiply all the rows because in the next lines we only select some of them
    inner_products = np.dot(sci_mean_sub_rows, np.require(kl_basis, requirements=['F']))
    # select the KLIP modes we want for each level of KLIP by multiplying by lower diagonal matrix
    lower_tri = np.tril(np.ones([max_basis, max_basis]))
    inner_products = inner_products * lower_tri
    # if there are NaNs due to negative eigenvalues, make sure they don't mess up the matrix multiplicatoin
    # by setting the appropriate values to zero
    if check_nans:
        needs_to_be_zeroed = np.where(lower_tri == 0)
        inner_products[needs_to_be_zeroed] = 0
        # make a KLIP PSF for each amount of klip basis, but only for the amounts of klip basis we actually output
        kl_basis[:, neg_evals] = 0
        klip_psf = np.dot(inner_products[numbasis,:], kl_basis.T)
        # for KLIP PSFs that use so many KL modes that they become nans, we have to put nan's back in those
        badbasis = np.where(numbasis >= np.min(neg_evals)) #use basis with negative eignevalues
        klip_psf[badbasis[0], :] = np.nan
    else:
        # make a KLIP PSF for each amount of klip basis, but only for the amounts of klip basis we actually output
        klip_psf = np.dot(inner_products[numbasis,:], kl_basis.T)

    # make subtracted image for each number of klip basis
    sub_img_rows_selected = sci_rows_selected - klip_psf

    # restore NaNs
    sub_img_rows_selected[sci_nanpix] = np.nan


    if return_basis is True:
        return sub_img_rows_selected.transpose(), kl_basis.transpose()
    elif return_basis_and_eig is True:
        return sub_img_rows_selected.transpose(), kl_basis.transpose(),evals*(np.size(sci)-1), evecs
    else:
        return sub_img_rows_selected.transpose()


    # old code that only did one number of KL basis for truncation
    # #truncation either based on user input or maximum number of PSFs
    # trunc_basis = np.min([numbasis, tot_basis])
    # #eigenvalues are ordered largest first now
    # eig_args = eig_args_all[0: trunc_basis]
    # kl_basis = kl_basis[:, eig_args]
    #
    # #project KL vectors onto science image to construct model PSF
    # inner_products = np.dot(sci_mean_sub, kl_basis)
    # klip_psf = np.dot(inner_products, kl_basis.T)
    #
    # #subtract from original image to get final image
    # sub_img = sci_mean_sub - klip_psf
    #
    # #restore NANs
    # sub_img[sci_nanpix] = np.nan
    #
    # #pdb.set_trace()
    #
    # return sub_img


def estimate_movement(radius, parang0=None, parangs=None, wavelength0=None, wavelengths=None, mode=None):
    """
    Estimates the movement of a hypothetical astrophysical source in ADI and/or SDI at the given radius and
    given reference parallactic angle (parang0) and reference wavelegnth (wavelength0)

    Args:
        radius: the radius from the star of the hypothetical astrophysical source
        parang0: the parallactic angle of the reference image (in degrees)
        parangs: array of length N of the parallactic angle of all N images (in degrees)
        wavelength0: the wavelength of the reference image
        wavelengths: array of length N of the wavelengths of all N images
        NOTE: we expect parang0 and parangs to be either both defined or both None.
                Same with wavelength0 and wavelengths
        mode: one of ['ADI', 'SDI', 'ADI+SDI'] for ADI, SDI, or ADI+SDI

    Returns:
        moves: array of length N of the distance an astrophysical source would have moved from the
               reference image
    """
    #default no movement parameters
    dtheta = 0 # how much the images moved in theta (polar coordinate)
    scale_fac = 1 # how much the image moved radially (r/radius)

    if (parang0 is not None):
        dtheta = np.radians(parang0 - parangs)
    if (wavelength0 is not None):
        scale_fac = (wavelength0/wavelengths)

    #define cartesean coordinate system where astrophysical source is at (x,y) = (r,0)
    x0 = radius
    y0 = 0.

    #find x,y location of astrophysical source for the rest of the images
    r = radius * scale_fac
    x = r * np.cos(dtheta)
    y = r * np.sin(dtheta)

    moves = np.sqrt((x-x0)**2 + (y-y0)**2)
    return moves


def calc_scaling(sats, refwv=18):
    """
    Helper function that calculates the wavelength scaling factor from the satellite spot locations.
    Uses the movement of spots diagonally across from each other, to calculate the scaling in a 
    (hopefully? tbd.) centering-independent way. 
    This method is definitely temporary and will be replaced by better scaling strategies as we come
    up with them.
    Scaling is calculated as the average of (1/2 * sqrt((x_1-x_2)**2+(y_1-y_2))), over the two pairs
    of spots.

    Args:
        sats: [4 x Nlambda x 2] array of x and y positions for the 4 satellite spots
        refwv: reference wavelength for scaling (optional, default = 20)
    Returns:
        scaling_factors: Nlambda array of scaling factors
    """
    pairs = [(0,3), (1,2)] # diagonally-located spots (spot_num - 1 for indexing)
    separations = np.mean([0.5*np.sqrt(np.diff(sats[p,:,0], axis=0)[0]**2 + np.diff(sats[p,:,1], axis=0)[0]**2) 
                           for p in pairs], 
                          axis=0) # average over each pair, the first axis

    scaling_factors = separations/separations[refwv]
    return scaling_factors


def nan_map_coordinates_2d(img, yp, xp, mc_kwargs=None):
    """
    scipy.ndimage.map_coordinates() that handles nans for 2-D transformations. Only works in 2-D!
    
    Do NaN detection by defining any pixel in the new coordiante system (xp, yp) as a nan
    If any one of the neighboring pixels in the original image is a nan (e.g. (xp, yp) = 
    (120.1, 200.1) is nan if either (120, 200), (121, 200), (120, 201), (121, 201) is a nan)

    Args:
        img (np.array): 2-D image that is looking to be transformed
        yp (np.array): 2-D array of y-coordinates that the image is evaluated out
        xp (np.array): 2-D array of x-coordinates that the image is evaluated out
        mc_kwargs (dict): other parameters to pass into the map_coordinates function.

    Returns:
        transformed_img (np.array): 2-D transformed image. Each pixel is evaluated at the (yp, xp) specified by xp and yp. 
    """
    # check if optional parameters are passed in
    if mc_kwargs is None:
        mc_kwargs = {}
    # if nothing specified, we will pad transformations with np.nan
    if "cval" not in mc_kwargs:
        mc_kwargs["cval"] = np.nan

    # check all four pixels around each pixel and see whether they are nans
    xp_floor = np.clip(np.floor(xp).astype(int), 0, img.shape[1]-1)
    xp_ceil = np.clip(np.ceil(xp).astype(int), 0, img.shape[1]-1)
    yp_floor = np.clip(np.floor(yp).astype(int), 0, img.shape[0]-1)
    yp_ceil = np.clip(np.ceil(yp).astype(int), 0, img.shape[0]-1)
    rotnans = np.where(np.isnan(img[yp_floor.ravel(), xp_floor.ravel()]) | 
                       np.isnan(img[yp_floor.ravel(), xp_ceil.ravel()]) |
                       np.isnan(img[yp_ceil.ravel(), xp_floor.ravel()]) |
                       np.isnan(img[yp_ceil.ravel(), xp_ceil.ravel()]))

    # resample image based on new coordinates, set nan values as median
    nanpix = np.where(np.isnan(img))
    medval = np.nanmedian(img)
    img_copy = np.copy(img)
    img_copy[nanpix] = medval
    transformed_img = ndimage.map_coordinates(img_copy, [yp, xp], **mc_kwargs)

    # mask nans
    img_shape = transformed_img.shape
    transformed_img.shape = [img_shape[0] * img_shape[1]]
    transformed_img[rotnans] = np.nan
    transformed_img.shape = img_shape

    return transformed_img


def align_and_scale(img, new_center, old_center=None, scale_factor=1, dtype=float):
    """
    Helper function that realigns and/or scales the image

    Args:
        img: 2D image to perform manipulation on
        new_center: 2 element tuple (xpos, ypos) of new image center
        old_center: 2 element tuple (xpos, ypos) of old image center
        scale_factor: how much the stretch/contract the image. Will we
                      scaled w.r.t the new_center (done after relaignment).
                      We will adopt the convention
                        >1: stretch image (shorter to longer wavelengths)
                        <1: contract the image (longer to shorter wvs)
                        This means scale factor should be lambda_0/lambda
                        where lambda_0 is the wavelength you want to scale to
    Returns:
        resampled_img: shifted and/or scaled 2D image
    """
    #create the coordinate system of the image to manipulate for the transform
    dims = img.shape
    x, y = np.meshgrid(np.arange(dims[1], dtype=dtype), np.arange(dims[0], dtype=dtype))
    mod_flag = 0 #check how many modifications we are making

    #if old_center is specified, realign the images
    if ((old_center is not None) & ~(np.array_equal(new_center, old_center))):
        dx = new_center[0] - old_center[0]
        dy = new_center[1] - old_center[1]
        x -= dx
        y -= dy
        mod_flag += 1
        #6/10/2016: Next line is a Bug fix
        new_center = old_center

    #if scale_factor is specified, scale the images
    if scale_factor != 1:
        #conver to polar for scaling
        r = np.sqrt((x - new_center[0]) ** 2 + (y - new_center[1]) ** 2)
        theta = np.arctan2(y - new_center[1], x - new_center[0])  #theta range is [-pi,pi]

        #Because x and y are the coordinates where we want to interpolate in the original image. See the following lines
        r /= scale_factor

        #convert back to cartesian
        x = r * np.cos(theta) + new_center[0]
        y = r * np.sin(theta) + new_center[1]
        mod_flag += 1

    #if nothing is to be changed, return a copy of the image
    if mod_flag == 0:
        return np.copy(img)

    resampled_img = nan_map_coordinates_2d(img, y, x)


    return resampled_img


def rotate(img, angle, center, new_center=None, flipx=False, astr_hdr=None):
    """
    Rotate an image by the given angle about the given center.
    Optional: can shift the image to a new image center after rotation. Also can reverse x axis for those left
              handed astronomy coordinate systems

    Args:
        img: a 2D image
        angle: angle CCW to rotate by (degrees)
        center: 2 element list [x,y] that defines the center to rotate the image to respect to
        new_center: 2 element list [x,y] that defines the new image center after rotation
        flipx: reverses x axis after rotation
        astr_hdr: wcs astrometry header for the image
    Returns:
        resampled_img: new 2D image
    """
    # skip this step if img is all nans
    if np.size(np.where(~np.isnan(img))) == 0:
        return np.copy(img)

    #convert angle to radians
    angle_rad = np.radians(angle)

    #create the coordinate system of the image to manipulate for the transform
    dims = img.shape
    x, y = np.meshgrid(np.arange(dims[1], dtype=np.float32), np.arange(dims[0], dtype=np.float32))

    #if necessary, move coordinates to new center
    if new_center is not None:
        dx = new_center[0] - center[0]
        dy = new_center[1] - center[1]
        x -= dx
        y -= dy

    #flip x if needed to get East left of North
    if flipx is True:
        x = center[0] - (x - center[0])

    #do rotation. CW rotation formula to get a CCW of the image
    xp = (x-center[0])*np.cos(angle_rad) + (y-center[1])*np.sin(angle_rad) + center[0]
    yp = -(x-center[0])*np.sin(angle_rad) + (y-center[1])*np.cos(angle_rad) + center[1]

    resampled_img = nan_map_coordinates_2d(img, yp, xp)

    #edit the astrometry header if given to compensate for orientation
    if astr_hdr is not None:
        _rotate_wcs_hdr(astr_hdr, angle, flipx=flipx)

    return resampled_img


def _rotate_wcs_hdr(wcs_header, rot_angle, flipx=False, flipy=False):
    """
    Modifies the wcs header when rotating/flipping an image.

    Args:
        wcs_header: wcs astrometry header
        rot_angle: in degrees CCW, the specified rotation desired
        flipx: after the rotation, reverse x axis? Yes if True
        flipy: after the rotation, reverse y axis? Yes if True
    """
    # rotate WCS header by a rotation matrix
    rot_angle_rad = np.radians(rot_angle)
    cos_rot = np.cos(rot_angle_rad)
    sin_rot = np.sin(rot_angle_rad)
    rot_matrix = np.array([[cos_rot, sin_rot], [-sin_rot, cos_rot]])
    wcs_header.wcs.cd = np.dot(wcs_header.wcs.cd, rot_matrix)

    # flip RA if true to be North up East left
    if flipx is True:
        wcs_header.wcs.cd[:,0] *= -1
    if flipy is True:
        wcs_header.wcs.cd[:,1] *= -1


def meas_contrast(dat, iwa, owa, resolution, center=None, low_pass_filter=True):
    """
    Measures the contrast in the image. Image must already be in contrast units and should be corrected for algorithm
    thoughput.

    Args:
        dat: 2D image - already flux calibrated
        iwa: inner working angle
        owa: outer working angle
        resolution: size of resolution element in pixels (FWHM or lambda/D)
        center: location of star (x,y). If None, defaults the image size // 2.
        low_pass_filter: if True, run a low pass filter.
                         Can also be a float which specifices the width of the Gaussian filter (sigma).
                         If False, no Gaussian filter is run

    Returns:
        (seps, contrast): tuple of separations in pixels and corresponding 5 sigma FPF

    """

    if center is None:
        starx = dat.shape[1]//2
        stary = dat.shape[0]//2
    else:
        starx, stary = center

    # figure out how finely to sample the radial profile
    dr = resolution/2.0
    numseps = int((owa-iwa)/dr)
    # don't want to start right at the edge of the occulting mask
    # but also want to well sample the contrast curve so go at twice the resolution
    seps = np.arange(numseps) * dr + iwa + resolution/2.0
    dsep = resolution
    # find equivalent Gaussian PSF for this resolution


    # run a low pass filter on the data, check if input is boolean or a number
    if not isinstance(low_pass_filter, bool):
        # manually passed in low pass filter size
        sigma = low_pass_filter
        filtered = nan_gaussian_filter(dat, sigma)
    elif low_pass_filter:
        # set low pass filter size to be same as resolution element
        sigma = dsep / 2.355  # assume resolution element size corresponds to FWHM
        filtered = nan_gaussian_filter(dat, sigma)
    else:
        # no filtering
        filtered = dat

    contrast = []
    # create a coordinate grid
    x,y = np.meshgrid(np.arange(float(dat.shape[1])), np.arange(float(dat.shape[0])))
    r = np.sqrt((x-starx)**2 + (y-stary)**2)
    theta = np.arctan2(y-stary, x-starx) % 2*np.pi
    for sep in seps:
        # calculate noise in an annulus with width of the resolution element
        annulus = np.where((r < sep + resolution/2) & (r > sep - resolution/2))
        noise_mean = np.nanmean(filtered[annulus])
        noise_std = np.nanstd(filtered[annulus], ddof=1)
        # account for small sample statistics
        num_samples = int(np.floor(2*np.pi*sep/resolution))

        # find 5 sigma flux using student-t statistics
        # Correction based on Mawet et al. 2014
        fpf_flux = t.ppf(0.99999971334, num_samples-1, scale=noise_std) * np.sqrt(1 + 1./num_samples) + noise_mean
        contrast.append(fpf_flux)

    return seps, np.array(contrast)


def nan_gaussian_filter(img, sigma):
    """
    Gaussian low-pass filter that handles nans

    Args:
        img: 2-D image
        sigma: float specifiying width of Gaussian

    Returns:
        filtered: 2-D image that has been smoothed with a Gaussian

    """
    # make a copy to mask with nans
    masked = np.copy(img)
    nan_locs = np.where(np.isnan(img))
    masked[nan_locs] = 0

    # filter the image
    filtered = ndimage.gaussian_filter(masked, sigma=sigma, truncate=4)

    # because of NaNs, we need to renormalize the gaussian filter, since NaNs shouldn't contribute
    norm_dat = np.ones(filtered.shape)
    norm_dat[nan_locs] = 0
    filter_norm = ndimage.gaussian_filter(norm_dat, sigma=sigma, truncate=4)
    filtered /= filter_norm
    filtered[nan_locs] = np.nan

    # for some reason, the fitlered image peak pixel fluxes get decreased by 2
    filtered *= 2

    return filtered


def high_pass_filter(img, filtersize=10):
    """
    A FFT implmentation of high pass filter.

    Args:
        img: a 2D image
        filtersize: size in Fourier space of the size of the space. In image space, size=img_size/filtersize

    Returns:
        filtered: the filtered image
    """
    # mask NaNs if there are any
    nan_index = np.where(np.isnan(img))
    if np.size(nan_index) > 0:
        good_index = np.where(~np.isnan(img))
        y, x = np.indices(img.shape)
        good_coords = np.array([x[good_index], y[good_index]]).T # shape of Npix, ndimage
        nan_fixer = sinterp.NearestNDInterpolator(good_coords, img[good_index])
        fixed_dat = nan_fixer(x[nan_index], y[nan_index])
        img[nan_index] = fixed_dat

    transform = fft.fft2(img)

    # coordinate system in FFT image
    u,v = np.meshgrid(fft.fftfreq(transform.shape[1]), fft.fftfreq(transform.shape[0]))
    # scale u,v so it has units of pixels in FFT space
    rho = np.sqrt((u*transform.shape[1])**2 + (v*transform.shape[0])**2)
    # scale rho up so that it has units of pixels in FFT space
    # rho *= transform.shape[0]
    # create the filter
    filt = 1. - np.exp(-(rho**2/filtersize**2))

    filtered = np.real(fft.ifft2(transform*filt))

    # restore NaNs
    filtered[nan_index] = np.nan
    img[nan_index] = np.nan

    return filtered


def define_annuli_bounds(annuli, IWA, OWA, annuli_spacing='constant'):
    """
    Defines the annuli boundaries radially

    Args:
        annuli: number of annuli
        IWA: inner working angle (pixels)
        OWA: outer working anglue (pixels)
        annuli_spacing: how to distribute the annuli radially. Currently three options. Constant (equally spaced),
                        log (logarithmical expansion with r), and linear (linearly expansion with r)

    Returns:
        rad_bounds: array of 2-element tuples that specify the beginning and end radius of that annulus

    """
    #calculate the annuli ranges
    if annuli_spacing.lower() == "constant":
        dr = float(OWA - IWA) / annuli
        rad_bounds = [(dr * rad + IWA, dr * (rad + 1) + IWA) for rad in range(annuli)]
    elif annuli_spacing.lower() == "log":
        # calculate normalization of log scaling
        unnormalized_log_scaling = np.log(np.arange(annuli) + 1) + 1
        log_coeff = float(OWA - IWA)/np.sum(unnormalized_log_scaling)
        # construct the radial spacing
        rad_bounds = []
        for i in range(annuli):
            # lower bound is either mask or end of previous annulus
            if i == 0:
                lower_bound = IWA
            else:
                lower_bound = rad_bounds[-1][1]
            upper_bound = lower_bound + log_coeff * unnormalized_log_scaling[i]
            rad_bounds.append((lower_bound, upper_bound))
    elif annuli_spacing.lower() == "linear":
        # scale linaer scaling to OWA-IWA
        linear_coeff = float(OWA - IWA)/np.sum(np.arange(annuli) + 1)
        rad_bounds = [(IWA + linear_coeff * rad, IWA + linear_coeff * (rad + 1)) for rad in range(annuli)]
    else:
        raise ValueError("annuli_spacing currently only supports 'constant', 'log', or 'linear'")

    # check to make sure the annuli are all greater than 1 pixel
    min_width = np.min(np.diff(rad_bounds, axis=1))
    if min_width < 1:
        raise ValueError("Too many annuli, some annuli are less than 1 pixel")

    return rad_bounds




__author__ = 'Jason Wang'
import numpy as np

def nMathar(wv, P, T, H=10):
    """
    Calculate the index of refraction as given by Mathar (2008): http://arxiv.org/pdf/physics/0610256v2.pdf
    ***Only valid for between 1.3 and 2.5 microns!

    Inputs:
        wv: wavelength in microns
        P:  Pressure in Pa
        T:  Temperature in Kelvin
        H:  relative humiditiy in % (i.e. between 0 and 100)
    Return:
        n:  index of refratoin
    """
    n = np.ones(np.size(wv)) # output. default to 1.
    wvnum = 1.e4/wv #cm^-1     # convert to wavenumbers

    # if it was passed in as a float, we need to convert it into a array for code reuse
    if not isinstance(wvnum, np.ndarray):
        wvnum = np.array([wvnum])

    # polynomial expansion in wavenumber
    # calcualate index of refraction by splitting it up by region
    # for region 1. < 2.65 microns, technically only valid for 1.3 to 2.5 microns
    region1 = np.where(wv < 2.8)
    if np.size(region1) > 0 :    
        wvnum0 = 1.e4/2.25 #cm^-1 #series expand around this wavenumber

        #do a 5th order expansion
        powers = np.arange(0,6)
        #calculate expansion coefficients
        coeffs = get_coeff_mathar(powers, P, T, H, 1)

        #sum of the power series expansion
        for coeff,power in zip(coeffs, powers):
            n[region1] += coeff * ((wvnum[region1] - wvnum0)**power)
    # next for everything greater than 2.65 microns. technically valid for 2.8 - 4.2 microns
    region2 = np.where(wv >= 2.8)
    if np.size(region2) > 0:
        wvnum0 = 1.e4/3.4 # cm^-1 #series expand around this wavenumber

        # do a 5th order expansion
        powers = np.arange(0,6)
        # calculate expansion coefficients
        coeffs = get_coeff_mathar(powers, P, T, H, 2)

        #sum of the power series expansion
        for coeff, power in zip(coeffs, powers):
            n[region2] += coeff * ((wvnum[region2] - wvnum0)**power)

    # return a int/float if that is what is passed in
    if isinstance(wv, (int, float)):
        n = n[0]

    return n

def get_coeff_mathar(i, P, T, H, wvrange=1):
    """
    Calculate the coefficients for the polynomial series expansion of index of refraction (Mathar (2008))
    ***Only valid for between 1.3 and 2.5 microns! and 2.8 through 4.2 microns!

    Inputs:
        i:  degree of expansion in wavenumber
        P:  Pressure in Pa
        T:  Temperature in Kelvin
        H:  relative humiditiy in % (i.e. between 0 and 100)
        wvrange (int): 1 = (1.3-2.5 um), 2 = (2.8 - 4.2 um)
    Return:
        coeff:  Coefficient [cm^-i]
    """

    #name all the constants in the model
    #series expansion in evironment parameters
    T0 = 273.15 + 17.5 #Kelvin
    P0 = 75000 #Pa
    H0 = 10 #%

    #delta terms for the expansion
    dT = 1./T - 1./T0
    dP = P - P0
    dH = H - H0

    #loads and loads of coefficients, see equation 7 in Mathar (2008)
    #use the power (i.e. i=[0..6]) to index the proper coefficient for that order
    if wvrange == 1:
        cref= np.array([ 0.200192e-3, 0.113474e-9, -0.424595e-14, 0.100957e-16,-0.293315e-20, 0.307228e-24]) # cm^i
        cT  = np.array([ 0.588625e-1,-0.385766e-7,  0.888019e-10,-0.567650e-13, 0.166615e-16,-0.174845e-20])  # K cm^i
        cTT = np.array([-3.01513,     0.406167e-3, -0.514544e-6,  0.343161e-9, -0.101189e-12, 0.106749e-16]) # K^2 cm^i
        cH  = np.array([-0.103945e-7, 0.136858e-11,-0.171039e-14, 0.112908e-17,-0.329925e-21, 0.344747e-25]) # cm^i / %
        cHH = np.array([ 0.573256e-12,0.186367e-16,-0.228150e-19, 0.150947e-22,-0.441214e-26, 0.461209e-30]) # cm^i / %^2
        cP  = np.array([ 0.267085e-8, 0.135941e-14, 0.135295e-18, 0.818218e-23,-0.222957e-26, 0.249964e-30]) # cm^i / Pa
        cPP = np.array([ 0.609186e-17,0.519024e-23,-0.419477e-27, 0.434120e-30,-0.122445e-33, 0.134816e-37]) # cm^i / Pa^2
        cTH = np.array([ 0.497859e-4,-0.661752e-8,  0.832034e-11,-0.551793e-14, 0.161899e-17,-0.169901e-21]) # cm^i K / %
        cTP = np.array([ 0.779176e-6, 0.396499e-12, 0.395114e-16, 0.233587e-20,-0.636441e-24, 0.716868e-28]) # cm^i K / Pa
        cHP = np.array([-0.206567e-15,0.106141e-20,-0.149982e-23, 0.984046e-27,-0.288266e-30, 0.299105e-34]) # cm^i / Pa %
    elif wvrange == 2:
        cref= np.array([ 0.200049e-3, 0.145221e-9,  0.250951e-12, -0.745834e-15,-0.161432e-17, 0.352780e-20]) # cm^i
        cT  = np.array([ 0.588431e-1,-0.825182e-7,  0.137982e-9,  0.352420e-13,-0.730651e-15,-0.167911e-18])  # K cm^i
        cTT = np.array([-3.13579,     0.694124e-3, -0.500604e-6, -0.116668e-8,  0.209644e-11, 0.591037e-14]) # K^2 cm^i
        cH  = np.array([-0.108142e-7, 0.230102e-11,-0.154652e-14,-0.323014e-17, 0.630616e-20, 0.173880e-22]) # cm^i / %
        cHH = np.array([ 0.586812e-12,0.312198e-16,-0.197792e-19,-0.461945e-22, 0.788398e-25, 0.245580e-27]) # cm^i / %^2
        cP  = np.array([ 0.266900e-8, 0.168162e-14, 0.353075e-17,-0.963455e-20,-0.223079e-22, 0.453166e-25]) # cm^i / Pa
        cPP = np.array([ 0.608860e-17,0.461560e-22, 0.184282e-24,-0.524471e-27,-0.121299e-29, 0.246512e-32]) # cm^i / Pa^2
        cTH = np.array([ 0.517962e-4,-0.112149e-7,  0.776507e-11, 0.172569e-13,-0.320582e-16,-0.899435e-19]) # cm^i K / %
        cTP = np.array([ 0.778638e-6, 0.446396e-12, 0.784600e-15,-0.195151e-17,-0.542083e-20, 0.103530e-22]) # cm^i K / Pa
        cHP = np.array([-0.217243e-15,0.104747e-20,-0.523689e-23, 0.817386e-26, 0.309913e-28,-0.363491e-31]) # cm^i / Pa %

    # use numpy arrays to calculate all the coefficients at the same time
    coeff = cref[i] + cT[i]*dT + cTT[i]*(dT**2) + cH[i]*dH + cHH[i]*(dH**2) + cP[i]*dP + cPP[i]*(dP**2) + cTH[i]*dT*dH + cTP[i]*dT*dP + cHP[i]*dH*dP

    return coeff


def nRoe(wv, P, T, H=10):
    """
    Compute n for air from the formula in Henry Roe's PASP paper: http://arxiv.org/pdf/astro-ph/0201273v1.pdf
    which in turn is referenced from Allen's Astrophysical Quantities.

    Inputs:
        wv: wavelength in microns
        P:  pressure in Pascal
        T:  temperature in Kelvin
        H:  relative humidity in % (0-100)
    Return:
        n:  index of refraction of air
    """

    #convert pressure from Pa to mbar
    P /= 100.

    #some constants in the function for n
    a1 = 64.328
    a2 = 29498.1
    a3 = 146.0
    a4 = 255.4
    a5 = 41.0

    Ts = 288.15   # K
    Ps = 1013.35 # mb

    #calculate n-1 for dry air
    K1 = 1e-6*(P/Ps * Ts/T)
    n1 = K1*(a1 + a2/(a3-wv**(-2)) + a4/(a5-wv**(-2))      )

    # water vapor correction
    # first compute partial pressure of water
    p_h20 = H/100. * saturation_pressure(T) # Pa
    fh20 = p_h20 / (1013.25 * 100)
    K2 = -43.49e-6 * fh20
    a6 = -7.956e-3
    nh2o = K2*(1 + a6*wv**(-2))

    return n1 + nh2o + 1

def saturation_pressure(temp):
    """
    Computes the saturation vapor pressure of water (from Voronin & Zheltikov 2017, eq 7)

    Args:
        temp (float): temperature in Kelvin
    
    Return:
        ps: saturation pressure in Pa
    """
    Tc = 647.096 # Kelvin, critical point temperature of water
    pc = 22.064e6 # Pa

    tau = Tc/temp
    theta = 1 - temp/Tc

    a1 = -7.85951783
    a2 = 1.84408259
    a3 = -11.7866497
    a4 = 22.6807411
    a5 = -15.9618719
    a6 = 1.80122502

    arg = a1*theta + a2*theta**1.5 + a3*theta**3 + a4*theta**3.5 + a5*theta**4 + a6*theta**7.5

    ps = pc * np.exp(tau * arg)

    return ps


import os, subprocess
import astropy.io.fits as fits
from astropy import wcs
import numpy as np
import scipy.ndimage as ndimage

from pyklip.instruments.Instrument import Data

class Ifs(Data):
    """
    A sequence of SPHERE IFS Data.

    Args:
        data_cube: FITS file with a 4D-cube (Nfiles, Nwvs, Ny, Nx) with all IFS coronagraphic data
            Also read spectral cubes and 2D images that have been saved using savedata().
        psf_cube: FITS file with a 3-D (Nwvs, Ny, Nx) PSF cube
        info_fits: FITS file with a table in the 1st ext hdr with parallactic angle info
        wavelenegth_info: FITS file with a 1-D array (Nwvs) of the wavelength sol'n of a cube
        psf_cube_size: size of the psf cube to save (length along 1 dimension)
        nan_mask_boxsize: size of box centered around any pixel <= 0 to mask as NaNs
        IWA: inner working angle of the data in arcsecs

    Attributes:
        input: Array of shape (N,y,x) for N images of shape (y,x)
        centers: Array of shape (N,2) for N centers in the format [x_cent, y_cent]
        filenums: Array of size N for the numerical index to map data to file that was passed in
        filenames: Array of size N for the actual filepath of the file that corresponds to the data
        ifs_rdp: Reduction algorithm used to obtain the input data.
        PAs: Array of N for the parallactic angle rotation of the target (used for ADI) [in degrees]
        wvs: Array of N wavelengths of the images (used for SDI) [in microns]. For polarization data, defaults to "None"
        IWA: a floating point scalar (not array). Specifies to inner working angle in pixels
        output: Array of shape (b, len(files), len(uniq_wvs), y, x) where b is the number of different KL basis cutoffs
        psfs: Spectral cube of size (Nwv, psfy, psfx) where psf_cube_size defines the size of psfy, psfx.
        psf_center: [x, y] location of the center of the PSF for a frame in self.psfs
        flipx: True by default. Determines whether a relfection about the x axis is necessary to rotate image North-up East left
        nfiles: number of datacubes
        nwvs: number of wavelengths

    """
    # class initialization
    # Astrometric calibration: Maire et al. 2016
    # North angle not used, images are only rotated by parallactic angle and pupil offset. True north correction needs to be applied to any astrometry
    # north_offset = -102.18 # (pupil offset + true north offset) who knows on the sign on this angle
    platescale = 0.007462

    # Coonstructor
    def __init__(self, data_cube, psf_cube, info_fits, wavelength_info, keepslices=None,
                 psf_cube_size=21, nan_mask_boxsize=9, IWA=0.15, object_name = None,
                 disable_minimum_filter = False, zeros2nans=False,subtract_psf_background=False):
        super(Ifs, self).__init__()

        # read in the data
        with fits.open(data_cube) as hdulist:
            self._input = hdulist[0].data # If 4D cube, Nfiles, Nwvs, Ny, Nx for vigan and Nwvs, Nfiles, Nx, Ny for sphere-dc
            # Read headers to be saved when using savedata. Vigan's or SPHERE-DC code don't include headers but pyklip does it
            # parameters or for the the location of injected planets.
            self.prihdrs = hdulist[0].header
            if np.size(self.input.shape) == 4:
                try:
                    self.prihdrs['HIERARCH ESO PRO REC1 ID'] # The SPHERE DC has headers with details of the reduction. We can use that information to differentiate between the two types of reduction
                except KeyError:
                    self.prihdrs['HIERARCH ESO PRO REC1 ID'] = False # If the keyword is missing, then we know it is data from Arthur Vigan's reduction
                if self.prihdrs['HIERARCH ESO PRO REC1 ID'] == 'sph_ifs_science_dr':
                    ifs_rdp = "sphere-dc" # Set the reduction process
                    self.input = np.swapaxes(self.input,0,1) # Swap the axes between the wavelengths and rotations for sphere-dc
                elif self.prihdrs['HIERARCH ESO PRO REC1 ID'] == False or self.prihdrs['HIERARCH ESO PRO REC1 ID'] == "F":
                    # determine which version of vigan pipeline it is, based on the info_fits file
                    # info_fits is a table for the old IDL pipeline, but a simple vector in the new Python pipeline
                    with fits.open(info_fits) as hdulist:
                        if len(hdulist) > 1:
                            # old IDL pipeline
                            ifs_rdp = "vigan-idl"     # Set the reduction process
                        else:
                            self.input = np.swapaxes(self.input,0,1) # Swap the axes between the wavelengths and rotations for vigan python pipeline
                            ifs_rdp = "vigan-python"  # Set the reduction process
                self._ifs_rdp = ifs_rdp # Store the reduction process
                self._filenums = np.repeat(np.arange(self.input.shape[0]), self.input.shape[1])
                self.nfiles = self.input.shape[0]
                self.nwvs = self.input.shape[1]
                # collapse files with wavelengths
                self.input = self.input.reshape(self.nfiles*self.nwvs, self.input.shape[2],
                                                self.input.shape[3])
                # zeros are nans, and anything adjacent to a pixel less than zero is 0.
                if not disable_minimum_filter:
                    input_minfilter = ndimage.minimum_filter(self.input, (0, nan_mask_boxsize, nan_mask_boxsize))
                    self.input[np.where(input_minfilter <= 0)] = np.nan
                if zeros2nans:
                    from scipy.signal import correlate2d
                    self.input[np.where(self.input == 0)] = np.nan
                    square = np.ones((nan_mask_boxsize,nan_mask_boxsize))
                    for k in range(self.input.shape[0]):
                        tmp_input = correlate2d(self.input[k,:,:],square,mode="same")
                        self.input[k,:,:][np.where(np.isnan(tmp_input))] = np.nan
                # centers are at dim/2
                self._centers = np.array([[img.shape[1]/2., img.shape[0]/2.] for img in self.input])
            elif np.size(self.input.shape) == 3:
                # If spectral data cube.
                self._filenums = np.zeros(self.input.shape[0])
                self.nfiles = 1
                self.nwvs = self.input.shape[0]
                # centers are at dim/2
                self.centers = np.array([[img.shape[1]/2., img.shape[0]/2.] for img in self.input])
            elif np.size(self.input.shape) == 2:
                # If 2D images like SNR maps.
                self._filenums = 0
                self.nfiles = 1
                self.nwvs = 1
                self.centers = np.array([[self.input.shape[1]/2., self.input.shape[0]/2.]])

        # read in the psf cube
        with fits.open(psf_cube) as hdulist:
            psf_cube = hdulist[0].data # Nwvs, Ny, Nx
            if np.size(psf_cube.shape) == 4: # If more than 1 PSF was taken during observation
                if ifs_rdp == "vigan-idl":
                    self.psfs = np.median(psf_cube, axis=0) # Take the median of the three PSFs
                elif ifs_rdp == "vigan-python":
                    self.psfs = np.median(psf_cube, axis=1) # Take the median of the three PSFs
                elif ifs_rdp == "sphere-dc":
                    self.psfs = np.median(psf_cube, axis=1) # Take the median of the three PSFs
            else:
                self.psfs = psf_cube
            self.psfs_center = [self.psfs.shape[2]/2., self.psfs.shape[1]/2.] # (x,y)

            # background subtraction
            if subtract_psf_background:
                tmp_size = 65
                pixelsbefore = tmp_size//2
                pixelsafter = tmp_size - pixelsbefore
                background_tmp = np.copy(self.psfs[:, int(self.psfs_center[1]-pixelsbefore):int(self.psfs_center[1]+pixelsafter),
                                                   int(self.psfs_center[0]-pixelsbefore):int(self.psfs_center[0]+pixelsafter)])
                x_grid, y_grid = np.meshgrid(np.arange(tmp_size)-tmp_size//2, np.arange(tmp_size)-tmp_size//2)
                r_grid = np.sqrt(x_grid**2 +y_grid**2)
                psf_mask = np.zeros((tmp_size,tmp_size))*np.nan
                psf_mask[np.where((20<r_grid)*(r_grid<30))] = 1
                background_tmp = background_tmp*psf_mask[None,:,:]

            # trim the cube
            pixelsbefore = psf_cube_size//2
            pixelsafter = psf_cube_size - pixelsbefore
            self.psfs = np.copy(self.psfs[:, int(self.psfs_center[1]-pixelsbefore):int(self.psfs_center[1]+pixelsafter),
                                          int(self.psfs_center[0]-pixelsbefore):int(self.psfs_center[0]+pixelsafter)])
            self.psfs_center = [psf_cube_size/2., psf_cube_size/2.]

            if subtract_psf_background:
                self.psfs = self.psfs - np.nanmedian(background_tmp,axis=(1,2))[:,None,None]

        # read in wavelength solution
        with fits.open(wavelength_info) as hdulist:
            self._wvs = hdulist[0].data
            # repeat for all Nfile cubes
            self._wvs = np.tile(self.wvs, self.nfiles)

        # read in PA info among other things. The two reduction processes handle PAs differently and therefore must be extracted differently
        with fits.open(info_fits) as hdulist:
            if ifs_rdp == "vigan-idl":
                metadata = hdulist[1].data
                self._PAs = np.repeat(metadata["PA"] + metadata['PUPOFF'], self.nwvs)
                self._filenames = np.repeat(metadata["FILE"], self.nwvs)
            elif ifs_rdp == "vigan-python":
                self._PAs = hdulist[0].data
                self._PAs = np.repeat(self.PAs, self.nwvs)
                self._filenames = np.repeat(data_cube, self.nwvs)
            elif ifs_rdp == "sphere-dc":
                self._PAs = -hdulist[0].data  # The SPHERE DC inverts the angles and we must correct for it
                self._PAs = np.repeat(self.PAs, self.nwvs)
                self._filenames = [data_cube]
                self._filenames = np.repeat(data_cube, self.nwvs)

        # we don't need to flip x for North Up East left
        self.flipx = False

        # I have no idea
        self.IWA = IWA / Ifs.platescale # 0.15" IWA

        # Creating WCS info for SPHERE
        self.wcs = []
        for vert_angle in self.PAs:
            w = wcs.WCS()
            vert_angle = np.radians(vert_angle)
            pc = np.array([[(-1)*np.cos(vert_angle), (-1)*-np.sin(vert_angle)],[np.sin(vert_angle), np.cos(vert_angle)]])
            cdmatrix = pc * self.platescale /3600.
            w.wcs.cd = cdmatrix
            self.wcs.append(w)
        self.wcs = np.array(self.wcs)
        # self.wcs = np.array([None for _ in range(self.nfiles * self.nwvs)])

        self._output = None

        # The definition of psfs_wvs requires that no wavelengths has been skipped in the input files
        # But it works with keepslices
        self.psfs_wvs = np.unique(self.wvs)
        self.star_peaks = np.nanmax(self.psfs,axis=(1,2))
        self.dn_per_contrast = np.squeeze(np.array([self.star_peaks[np.where(self.psfs_wvs==wv)[0]] for wv in self.wvs]))

        if np.size(self.input.shape) == 2:
            self.wvs = 0
            self.dn_per_contrast = 0

        if keepslices is not None:
            self.input = self.input[keepslices,:,:]
            self.nfiles = self.input.shape[0]//self.nwvs # only works if keepslices select whole cubes
            self.nwvs = self.input.shape[0]//self.nfiles # only works if keepslices select whole cubes
            self.filenums = self.filenums[keepslices]
            self.centers = self.centers[keepslices]
            self.wvs = self.wvs[keepslices]
            self.PAs = self.PAs[keepslices]
            self.filenames = self.filenames[keepslices]
            self.dn_per_contrast = self.dn_per_contrast[keepslices]
            self.wcs = self.wcs[keepslices]

        # Required for automatically querying Simbad for the spectral type of the star.
        self.object_name = object_name
        # self.object_name = os.path.basename(data_cube).split("-")[0].split("_")[0]

    ################################
    ### Instance Required Fields ###
    ################################

    @property
    def input(self):
        return self._input
    @input.setter
    def input(self, newval):
        self._input = newval

    @property
    def ifs_rdp(self):
        return self._ifs_rdp
    @ifs_rdp.setter
    def ifs_rdp(self, newval):
        self._ifs_rdp = newval

    @property
    def centers(self):
        return self._centers
    @centers.setter
    def centers(self, newval):
        self._centers = newval

    @property
    def filenums(self):
        return self._filenums
    @filenums.setter
    def filenums(self, newval):
        self._filenums = newval

    @property
    def filenames(self):
        return self._filenames
    @filenames.setter
    def filenames(self, newval):
        self._filenames = newval

    @property
    def PAs(self):
        return self._PAs
    @PAs.setter
    def PAs(self, newval):
        self._PAs = newval

    @property
    def wvs(self):
        return self._wvs
    @wvs.setter
    def wvs(self, newval):
        self._wvs = newval

    @property
    def wcs(self):
        return self._wcs
    @wcs.setter
    def wcs(self, newval):
        self._wcs = newval

    @property
    def IWA(self):
        return self._IWA
    @IWA.setter
    def IWA(self, newval):
        self._IWA = newval

    @property
    def output(self):
        return self._output
    @output.setter
    def output(self, newval):
        self._output = newval


    ###############
    ### Methods ###
    ###############

    def readdata(self, filepaths):
        """
        Reads in the data from the files in the filelist and writes them to fields
        """
        pass


    def savedata(self, filepath, data, klipparams=None, filetype="", zaxis=None , more_keywords=None):
        """
        Save SPHERE Data.

        Note: In principle, the function only works inside klip_dataset(). In order to use it outside of klip_dataset,
            you need to define the follwing attributes:
                dataset.output_centers = dataset.centers

        Args:
            filepath: path to file to output
            data: 2D or 3D data to save
            klipparams: a string of klip parameters
            filetype: filetype of the object (e.g. "KL Mode Cube", "PSF Subtracted Spectral Cube")
            zaxis: a list of values for the zaxis of the datacub (for KL mode cubes currently)
            more_keywords (dictionary) : a dictionary {key: value, key:value} of header keywords and values which will
                                         written into the primary header

        """
        hdulist = fits.HDUList()
        hdulist.append(fits.PrimaryHDU(data=data,header=self.prihdrs))

        # save all the files we used in the reduction
        # we'll assume you used all the input files
        # remove duplicates from list
        if self.ifs_rdp == "vigan-idl" or self.ifs_rdp == "vigan-python":
            filenames = np.unique(self.filenames)
            nfiles = np.size(filenames)
            hdulist[0].header["DRPNFILE"] = (nfiles, "Num raw files used in pyKLIP")
            for i, filename in enumerate(filenames):
                hdulist[0].header["FILE_{0}".format(i)] = filename + '.fits'

        # write out psf subtraction parameters
        # get pyKLIP revision number
        pykliproot = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        # the universal_newline argument is just so python3 returns a string instead of bytes
        # this will probably come to bite me later
        try:
            pyklipver = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD'], cwd=pykliproot, universal_newlines=True).strip()
        except:
            pyklipver = "unknown"
        hdulist[0].header['PSFSUB'] = ("pyKLIP", "PSF Subtraction Algo")
        hdulist[0].header.add_history("Reduced with pyKLIP using commit {0}".format(pyklipver))
        hdulist[0].header['CREATOR'] = "pyKLIP-{0}".format(pyklipver)

        # store commit number for pyklip
        hdulist[0].header['pyklipv'] = (pyklipver, "pyKLIP version that was used")

        if klipparams is not None:
            hdulist[0].header['PSFPARAM'] = (klipparams, "KLIP parameters")
            hdulist[0].header.add_history("pyKLIP reduction with parameters {0}".format(klipparams))

        # store extra keywords in header
        if more_keywords is not None:
            for hdr_key in more_keywords:
                hdulist[0].header[hdr_key] = more_keywords[hdr_key]

        # write z axis units if necessary
        if zaxis is not None:
            # Writing a KL mode Cube
            if "KL Mode" in filetype:
                hdulist[0].header['CTYPE3'] = 'KLMODES'
                # write them individually
                for i, klmode in enumerate(zaxis):
                    hdulist[0].header['KLMODE{0}'.format(i)] = (klmode, "KL Mode of slice {0}".format(i))
                hdulist[0].header['CUNIT3'] = "N/A"
                hdulist[0].header['CRVAL3'] = 1
                hdulist[0].header['CRPIX3'] = 1.
                hdulist[0].header['CD3_3'] = 1.

        if "Spectral" in filetype:
            uniquewvs = np.unique(self.wvs)
            # do spectral stuff instead
            # because wavelength solutoin is nonlinear, we're not going to store it here
            hdulist[0].header['CTYPE3'] = 'WAVE'
            hdulist[0].header['CUNIT3'] = "N/A"
            hdulist[0].header['CRPIX3'] = 1.
            hdulist[0].header['CRVAL3'] = 0
            hdulist[0].header['CD3_3'] = 1
            # write it out instead
            for i, wv in enumerate(uniquewvs):
                hdulist[0].header['WV{0}'.format(i)] = (wv, "Wavelength of slice {0}".format(i))

        center = self.output_centers[0]
            
        hdulist[0].header.update({'PSFCENTX': center[0], 'PSFCENTY': center[1]})
        hdulist[0].header.update({'CRPIX1': center[0], 'CRPIX2': center[1]})
        hdulist[0].header.add_history("Image recentered to {0}".format(str(center)))

        try:
            hdulist.writeto(filepath, overwrite=True)
        except TypeError:
            hdulist.writeto(filepath, clobber=True)
        hdulist.close()

    def calibrate_output(self, img, spectral=False, units="contrast"):
        """
        Calibrates the flux of an output image. Can either be a broadband image or a spectral cube depending
        on if the spectral flag is set.

        Args:
            img: unclaibrated image.
                 If spectral is not set, this can either be a 2-D or 3-D broadband image
                 where the last two dimensions are [y,x]
                 If specetral is True, this is a 3-D spectral cube with shape [wv,y,x]
            spectral: if True, this is a spectral datacube. Otherwise, it is a broadband image.
            units: currently only support "contrast" w.r.t central star

        Return:
            img: calibrated image of the same shape (this is the same object as the input!!!)
        """
        if units == "contrast":
            if spectral:
                # spectral cube, each slice needs it's own calibration
                numwvs = img.shape[0]
                print(self.dn_per_contrast.shape)
                print(img.shape)
                img /= self.dn_per_contrast[:numwvs, None, None]
            else:
                # broadband image
                img /= np.nanmean(self.dn_per_contrast)
            self.flux_units = "contrast"

        return img


class Irdis(Data):
    """
    A sequence of SPHERE IRDIS Data.

    Args:
        data_cube: FITS file with a 4D-cube (Nfiles, Nwvs, Ny, Nx) with all IFS coronagraphic data
            Also read spectral cubes and 2D images that have been saved using savedata().
        psf_cube: FITS file with a 3-D (Nwvs, Ny, Nx) PSF cube
            If None, psf_cube = data_cube.replace("cube_coro","cube_psf")
        info_fits: FITS file with a table in the 1st ext hdr with parallactic angle info
            If None, info_fits = data_cube.replace("cube_coro","info")
        wavelength_str: string to specifiy the band (e.g. "H2H3", "K1K2")
        psf_cube_size: size of the psf cube to save (length along 1 dimension)
        IWA: inner working angle of the data in arcsecs
        OWA: outer working angle of the data in arcsecs

    Attributes:
        input: Array of shape (N,y,x) for N images of shape (y,x)
        centers: Array of shape (N,2) for N centers in the format [x_cent, y_cent]
        filenums: Array of size N for the numerical index to map data to file that was passed in
        filenames: Array of size N for the actual filepath of the file that corresponds to the data
        PAs: Array of N for the parallactic angle rotation of the target (used for ADI) [in degrees]
        wvs: Array of N wavelengths of the images (used for SDI) [in microns]. For polarization data, defaults to "None"
        IWA: a floating point scalar (not array). Specifies to inner working angle in pixels
        OWA: a floating point scalar (not array). Specifies to out working angle in pixels
        output: Array of shape (b, len(files), len(uniq_wvs), y, x) where b is the number of different KL basis cutoffs
        psfs: Spectral cube of size (2, psfy, psfx) where psf_cube_size defines the size of psfy, psfx.
        psf_center: [x, y] location of the center of the PSF for a frame in self.psfs
        flipx: True by default. Determines whether a relfection about the x axis is necessary to rotate image North-up East left
        nfiles: number of datacubes
        prihdrs: SPHERE headers if reduced by Sphere data center
        nwvs: number of wavelengths (i.e. 2 for dual band imaging)
    """
    # class initialization
    # Astrometric calibration: Maire et al. 2016
    # North angle not used, images are only rotated by parallactic angle and pupil offset. True north correction needs to be applied to any astrometry
    # north_offset = -1.75 # who knows on the sign on this angle
    platescale = 0.012255
    
    # dual band imaging central wavelengths
    wavelengths = {"Y2Y3": (1.022, 1.076), "J2J3": (1.190, 1.273), "H2H3": (1.593, 1.667),
                   "H3H4": (1.667, 1.733), "K1K2": (2.110, 2.251),
                   "B_Y": (1.043, 1.043), "B_J": (1.245, 1.245), "B_H": (1.625, 1.625),
                   "B_Ks": (2.182, 2.182)}

    # Coonstructor
    def __init__(self, data_cube, psf_cube, info_fits, wavelength_str, psf_cube_size=21, IWA=0.08, OWA=None,
                 keepslices=None):
        super(Irdis, self).__init__()

        # read in the data
        with fits.open(data_cube) as hdulist:
            self._input = hdulist[0].data # 4D cube, Nfiles, Nwvs, Ny, Nx
            # Read headers to be saved when using savedata. Vigan's code doesn't include headers but pyklip does it
            # parameters or for the the location of injected planets.
            self._filenames = [data_cube]
            self.prihdrs = hdulist[0].header
            if np.size(self.input.shape) == 4: # If 4D
                if ('PIXSCAL' in self.prihdrs):
                    irdis_rdp = "sphere-dc" # Set the reduction process
                    self.input = np.swapaxes(self.input,0,1) # Swap the axes between the wavelengths and rotations for sphere-dc
                else:
                    with fits.open(info_fits) as hdulist:
                        if len(hdulist) > 1:
                            # old IDL pipeline
                            irdis_rdp = "vigan-idl"     # Set the reduction process
                        else:
                            irdis_rdp = "vigan-python" # Set the reduction process
                            self.input = np.swapaxes(self.input,0,1) # Swap the axes between the wavelengths and rotations for vigan python pipeline
                self.irdis_rdp = irdis_rdp # Store the reduction process
                self._filenums = np.repeat(np.arange(self.input.shape[0]), self.input.shape[1])
                self.nfiles = self.input.shape[0]
                self.nwvs = self.input.shape[1]
                # collapse files with wavelengths
                self.input = self.input.reshape(self.nfiles*self.nwvs, self.input.shape[2],
                                                self.input.shape[3])
                if OWA is not None:
                    # Trim cube to region of interest to speed up processing
                    if (OWA / Irdis.platescale) < (0.45*self.input.shape[1]):
                        # Only do it if the OWA in pixels is a bit less than half the image size
                        trim_px = int((self.input.shape[1]/2.0) - (OWA / Irdis.platescale))
                        self.input = self.input[:, trim_px:-trim_px, trim_px:-trim_px]

                # centers are at dim/2
                self._centers = np.array([[img.shape[1]/2., img.shape[0]/2.] for img in self.input])
            elif np.size(self.input.shape) == 3:
                # If spectral data cube.
                self._filenums = np.zeros(self.input.shape[0])
                self.nfiles = 1
                self.nwvs = self.input.shape[0]
                # centers are at dim/2
                self.centers = np.array([[img.shape[1]/2., img.shape[0]/2.] for img in self.input])
            elif np.size(self.input.shape) == 2:
                # If 2D images like SNR maps.
                self._filenums = 0
                self.nfiles = 1
                self.nwvs = 1
                self.centers = np.array([[self.input.shape[1]/2., self.input.shape[0]/2.]])

        # read in the psf cube
        with fits.open(psf_cube) as hdulist:
            psf_cube = hdulist[0].data # Nwvs, Ny, Nx
            if np.size(psf_cube.shape) == 4:
                if irdis_rdp == "vigan-idl":
                    self.psfs = np.median(psf_cube, axis=0) 
                elif irdis_rdp == "vigan-python":
                    self.psfs = np.median(psf_cube, axis=1) # Take the median of the three PSFs
                elif irdis_rdp == "sphere-dc":
                    self.psfs = np.median(psf_cube, axis=1) # Take the median of the three PSFs
            else:
                self.psfs = psf_cube
                # multiple PSF sequences were taken. Collapse them and take the average
            self.psfs_center = [self.psfs.shape[2]/2., self.psfs.shape[1]/2.] # (x,y)

            # trim the cube
            pixelsbefore = psf_cube_size//2
            pixelsafter = psf_cube_size - pixelsbefore
            self.psfs = np.copy(self.psfs[:, int(self.psfs_center[1]-pixelsbefore):int(self.psfs_center[1]+pixelsafter),
                                          int(self.psfs_center[0]-pixelsbefore):int(self.psfs_center[0]+pixelsafter)])
            self.psfs_center = [psf_cube_size/2., psf_cube_size/2.]

        db_wvs = Irdis.wavelengths[wavelength_str]
        self._wvs = np.tile(db_wvs, self.nfiles)

        # read in PA info among other things
        with fits.open(info_fits) as hdulist:
            if irdis_rdp == "vigan-idl":
                metadata = hdulist[1].data
                self._PAs = np.repeat(metadata["PA"] + metadata['PUPOFF'], self.nwvs)
                self._filenames = np.repeat(metadata["FILE"], self.nwvs)
            elif irdis_rdp == "vigan-python":
                self._PAs = hdulist[0].data
                self._PAs = np.repeat(self.PAs, self.nwvs)
                self._filenames = np.repeat(data_cube, self.nwvs)
            elif irdis_rdp == "sphere-dc":
                self._PAs = -hdulist[0].data # The SPHERE DC inverts the angles and we must correct for it
                self._PAs = np.repeat(self.PAs, self.nwvs)
                self._filenames = np.repeat(data_cube, self.nwvs)

        # we don't need to flip x for North Up East left
        self.flipx = False

        # I have no idea
        self.IWA = IWA / Irdis.platescale # 0.2" IWA
        self.OWA = OWA / Irdis.platescale if OWA is not None else None

        # Creating WCS info for SPHERE
        self.wcs = []
        for vert_angle in self.PAs:
            w = wcs.WCS()
            vert_angle = np.radians(vert_angle)
            pc = np.array([[(-1)*np.cos(vert_angle), (-1)*-np.sin(vert_angle)],[np.sin(vert_angle), np.cos(vert_angle)]])
            cdmatrix = pc * self.platescale /3600.
            w.wcs.cd = cdmatrix
            self.wcs.append(w)
        self.wcs = np.array(self.wcs)
        # self.wcs = np.array([None for _ in range(self.nfiles * self.nwvs)])

        self._output = None

        # The definition of psfs_wvs requires that no wavelengths has been skipped in the input files
        # But it works with keepslices
        self.psfs_wvs = np.unique(self.wvs)
        self.star_peaks = np.nanmax(self.psfs,axis=(1,2))
        self.dn_per_contrast = np.squeeze(np.array([self.star_peaks[np.where(self.psfs_wvs==wv)[0]] for wv in self.wvs]))

        if np.size(self.input.shape) == 2:
            self.wvs = 0
            self.dn_per_contrast = 0

        if keepslices is not None:
            self.input = self.input[keepslices,:,:]
            self.nfiles = self.input.shape[0]//self.nwvs # only works if keepslices select whole cubes
            self.nwvs = self.input.shape[0]//self.nfiles # only works if keepslices select whole cubes
            self.filenums = self.filenums[keepslices]
            self.centers = self.centers[keepslices]
            self.wvs = self.wvs[keepslices]
            self.PAs = self.PAs[keepslices]
            self.filenames = self.filenames[keepslices]
            self.dn_per_contrast = self.dn_per_contrast[keepslices]
            self.wcs = self.wcs[keepslices]

        # Required for automatically querying Simbad for the spectral type of the star.
        self.object_name = os.path.basename(data_cube).split("_")[0]

    ################################
    ### Instance Required Fields ###
    ################################

    @property
    def input(self):
        return self._input
    @input.setter
    def input(self, newval):
        self._input = newval

    @property
    def centers(self):
        return self._centers
    @centers.setter
    def centers(self, newval):
        self._centers = newval

    @property
    def filenums(self):
        return self._filenums
    @filenums.setter
    def filenums(self, newval):
        self._filenums = newval

    @property
    def filenames(self):
        return self._filenames
    @filenames.setter
    def filenames(self, newval):
        self._filenames = newval

    @property
    def PAs(self):
        return self._PAs
    @PAs.setter
    def PAs(self, newval):
        self._PAs = newval

    @property
    def wvs(self):
        return self._wvs
    @wvs.setter
    def wvs(self, newval):
        self._wvs = newval

    @property
    def wcs(self):
        return self._wcs
    @wcs.setter
    def wcs(self, newval):
        self._wcs = newval

    @property
    def IWA(self):
        return self._IWA
    @IWA.setter
    def IWA(self, newval):
        self._IWA = newval

    @property
    def OWA(self):
        return self._OWA
    @OWA.setter
    def OWA(self, newval):
        self._OWA = newval

    @property
    def output(self):
        return self._output
    @output.setter
    def output(self, newval):
        self._output = newval


    ###############
    ### Methods ###
    ###############

    def readdata(self, filepaths):
        """
        Reads in the data from the files in the filelist and writes them to fields
        """
        pass


    def savedata(self, filepath, data, klipparams=None, filetype="", zaxis=None , more_keywords=None):
        """
        Save SPHERE Data.

        Note: In principle, the function only works inside klip_dataset(). In order to use it outside of klip_dataset,
            you need to define the follwing attribute:
                dataset.output_centers = dataset.centers

        Args:
            filepath: path to file to output
            data: 2D or 3D data to save
            klipparams: a string of klip parameters
            filetype: filetype of the object (e.g. "KL Mode Cube", "PSF Subtracted Spectral Cube")
            zaxis: a list of values for the zaxis of the datacub (for KL mode cubes currently)
            more_keywords (dictionary) : a dictionary {key: value, key:value} of header keywords and values which will
                                         written into the primary header

        """
        hdulist = fits.HDUList()
        hdulist.append(fits.PrimaryHDU(data=data))

        # save all the files we used in the reduction
        # we'll assume you used all the input files
        # remove duplicates from list
        if self.irdis_rdp == "vigan-python":
            filenames = np.unique(self.filenames)
            nfiles = np.size(filenames)
            hdulist[0].header["DRPNFILE"] = (nfiles, "Num raw files used in pyKLIP")
            for i, filename in enumerate(filenames):
                hdulist[0].header["FILE_{0}".format(i)] = filename + '.fits'

        # write out psf subtraction parameters
        # get pyKLIP revision number
        pykliproot = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        # the universal_newline argument is just so python3 returns a string instead of bytes
        # this will probably come to bite me later
        try:
            pyklipver = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD'], cwd=pykliproot, universal_newlines=True).strip()
        except:
            pyklipver = "unknown"
        hdulist[0].header['PSFSUB'] = ("pyKLIP", "PSF Subtraction Algo")
        hdulist[0].header.add_history("Reduced with pyKLIP using commit {0}".format(pyklipver))
        hdulist[0].header['CREATOR'] = "pyKLIP-{0}".format(pyklipver)

        # store commit number for pyklip
        hdulist[0].header['pyklipv'] = (pyklipver, "pyKLIP version that was used")

        if klipparams is not None:
            hdulist[0].header['PSFPARAM'] = (klipparams, "KLIP parameters")
            hdulist[0].header.add_history("pyKLIP reduction with parameters {0}".format(klipparams))

        # store extra keywords in header
        if more_keywords is not None:
            for hdr_key in more_keywords:
                hdulist[0].header[hdr_key] = more_keywords[hdr_key]

        # write z axis units if necessary
        if zaxis is not None:
            # Writing a KL mode Cube
            if "KL Mode" in filetype:
                hdulist[0].header['CTYPE3'] = 'KLMODES'
                # write them individually
                for i, klmode in enumerate(zaxis):
                    hdulist[0].header['KLMODE{0}'.format(i)] = (klmode, "KL Mode of slice {0}".format(i))
                hdulist[0].header['CUNIT3'] = "N/A"
                hdulist[0].header['CRVAL3'] = 1
                hdulist[0].header['CRPIX3'] = 1.
                hdulist[0].header['CD3_3'] = 1.

        if "Spectral" in filetype:
            uniquewvs = np.sort(np.unique(self.wvs))
            # do spectral stuff instead
            # because wavelength solutoin is nonlinear, we're not going to store it here
            hdulist[0].header['CTYPE3'] = 'WAVE'
            hdulist[0].header['CUNIT3'] = "microns"
            hdulist[0].header['CRPIX3'] = 1.
            # there's only 2 wvs in DBI
            hdulist[0].header['CRVAL3'] = uniquewvs[0]
            hdulist[0].header['CD3_3'] = uniquewvs[1] - uniquewvs[0]
            # write it out instead

        center = self.output_centers[0]

        hdulist[0].header.update({'PSFCENTX': center[0], 'PSFCENTY': center[1]})
        hdulist[0].header.update({'CRPIX1': center[0], 'CRPIX2': center[1]})
        hdulist[0].header.add_history("Image recentered to {0}".format(str(center)))

        try:
            hdulist.writeto(filepath, overwrite=True)
        except TypeError:
            hdulist.writeto(filepath, clobber=True)
        hdulist.close()

    def calibrate_output(self, img, spectral=False, units="contrast"):
        """
        Calibrates the flux of an output image. Can either be a broadband image or a spectral cube depending
        on if the spectral flag is set.

        Args:
            img: unclaibrated image.
                 If spectral is not set, this can either be a 2-D or 3-D broadband image
                 where the last two dimensions are [y,x]
                 If specetral is True, this is a 3-D spectral cube with shape [wv,y,x]
            spectral: if True, this is a spectral datacube. Otherwise, it is a broadband image.
            units: currently only support "contrast" w.r.t central star

        Return:
            img: calibrated image of the same shape (this is the same object as the input!!!)
        """
        # return img
        if units == "contrast":
            if spectral:
                # spectral cube, each slice needs it's own calibration
                numwvs = img.shape[0]
                img /= self.dn_per_contrast[:numwvs, None, None]
            else:
                # broadband image
                img /= np.nanmean(self.dn_per_contrast)
            self.flux_units = "contrast"

        return img



#!/usr/bin/env python
"""
Given a datacube, find the four corresponding spot files. 
Plot the calculated positions on top of the original cube.

Run from an ipython terminal with:
%run spot_checker.py full/path/to/cube.fits
"""

from __future__ import division

import sys
import os
import warnings
import glob

from multiprocessing import Pool, Process, Queue

import numpy as np

import matplotlib as mpl

import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
from matplotlib.colors import LogNorm
from matplotlib.collections import PatchCollection
from matplotlib.patches import CirclePolygon

from astropy.io import fits
from astropy import units

import argparse
#for handling different python versions
if sys.version_info < (3,0):
    import ConfigParser
else:
    import configparser as ConfigParser

# needed to import from outside this folder
base_dir = os.path.dirname(__file__) or '.'
sys.path.append(base_dir)
import pyklip.instruments.P1640_support.P1640spots

dnah_spot_directory = '/data/p1640/data/users/spot_positions/jonathan/'


# plt.ion()


# use multiple threads - one for drawing the figure, and another for handling user input

"""
Pseudocode:
1. Load list of files
2. Create the "good files" dictionary
3. For each file:
3a. Split offt a thread for drawing the cube
3b. Ask for user input
4. When the user provides 'y' or 'n', update the dictionary and kill the drawing thread
5. Move on to the next file
"""

def get_total_exposure_time(fitsfiles, unit=units.minute):
    """
    Accept a list of fits files and return the total exposure time
    Input:
      fitsfiles: single fits file or list of files with keyword 'EXPTIME' in the header
      units: [minute] astropy.units unit for the output
    Output:
      tot_exp_time: the sum of the exposure times for each cube, in minutes
    """
    exptimes = np.array([fits.getval(f, 'EXP_TIME') for f in fitsfiles]) * units.second
    return np.sum(exptimes).to(unit)


# open a fits file and draw the cube
def draw_cube(cube, cube_name, header, seeing, airmass, cube_ix):
    """
    Make a figure and draw cube slices on it
    """
    chan = 14
    nchan = cube.shape[0]
    max_val = np.max(cube)
    fig = plt.figure()
    fig.suptitle(cube_name, fontsize='x-large')
    gridsize=(8,8)
    
    ax_seeing = plt.subplot2grid(gridsize, (4,5), rowspan=2, colspan=3)
    ax_seeing.plot(seeing, 'b')
    ax_seeing.set_xlim(xmin=-0.5, xmax=len(seeing)+0.5)
    ax_seeing.set_yticks(np.sort(seeing)[[0,-1]])
    ax_seeing.axvline(cube_ix, c='k', ls='--')
    ax_seeing.axhline(seeing[cube_ix], c='k', ls='--')
    ax_seeing.yaxis.set_major_formatter(mticker.FormatStrFormatter('%.2f'))
    ax_seeing.set_title("seeing")
    ax_airmass = plt.subplot2grid(gridsize, (6,5), rowspan=2, colspan=3)
    ax_airmass.plot(airmass, 'r')
    ax_airmass.set_xlim(xmin=-0.5, xmax=len(airmass)+0.5)
    ax_airmass.set_yticks(np.sort(airmass)[[0,-1]])
    ax_airmass.axvline(cube_ix, c='k', ls='--')
    ax_airmass.axhline(airmass[cube_ix], c='k', ls='--')
    ax_airmass.yaxis.set_major_formatter(mticker.FormatStrFormatter('%.2f'))
    ax_airmass.set_title("airmass")
    ax_airmass.set_xlabel("cube #", size='large')

    #ax = fig.add_subplot(111)
    ax = plt.subplot2grid(gridsize, (1,0),
                          rowspan=6,
                          colspan=5)
#    ax.yaxis.tick_right()
#    fig.subplots_adjust(left=0.3)
    datainfo = """Exp time: {exptime:.3f}
Seeing: {seeing:>9.3f}
Airmass: {airmass:>8.3f}
Max val: {maxval:>8.1f}
Min val: {minval:>8.1f}""".format(exptime=header["EXP_TIME"],
                              seeing=header["SEEING"],
                              airmass=header["INIT_AM"],
                              maxval=np.nanmax(cube),
                              minval=np.nanmin(cube))
    
    fig.text(0.67, 0.85, datainfo, size='large', family='monospace',
             linespacing=2, 
             verticalalignment='top')

    plt.draw()
    plt.tight_layout(pad=1)
    while True:
        #plt.sca(ax)
        #plt.cla()
        ax.clear()
        chan = chan % nchan
        # you're gonna think you want a common scale for all the slices but you're wrong, leave LogNorm alone
        imax = ax.imshow(cube[chan], norm=LogNorm())#vmax=max_val))
        #ax.set_title("{name}\nChannel {ch:02d}".format(name=cube_name, ch=chan))
        ax.set_title("Channel {ch:02d}".format(name=cube_name, ch=chan))
        plt.pause(0.25)
        chan += 1

def plot_airmass_and_seeing(fitsfiles):
    # plot airmass and seeing
    seeing = [fits.getval(ff,'SEEING') for ff in fitsfiles]
    airmass = [fits.getval(ff,'INIT_AM') for ff in fitsfiles]
    fig, axes = plt.subplots(nrows=1, ncols=2)
    axes[0].plot(seeing,'b-')
    axes[0].axhline(np.mean(seeing), c='b',label='mean seeing')
    axes[0].set_title("Seeing")
    axes[1].plot(airmass, 'r-')
    axes[1].axhline(np.mean(airmass), c='r', label='mean airmass')
    axes[1].set_title("Airmass")
    plt.draw()

    

def run_checker(fitsfiles):
    """
    Run the checker
    """
    if fitsfiles == 'help':
        usage()
        return
    seeing = [fits.getval(ff,'SEEING') for ff in fitsfiles]
    airmass = [fits.getval(ff,'INIT_AM') for ff in fitsfiles]
    
    good_cubes = dict(zip(fitsfiles, [None for f in fitsfiles]))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        #fig = plt.figure()
        repeat = True
        while repeat:
#            proc_cube_stats = Process(target=plot_airmass_and_seeing,
#                                         args=[fitsfiles])
#            proc_cube_stats.start()
            for i, ff in enumerate(fitsfiles):
                # check file
                if not os.path.isfile(ff):
                    print("File not found: {0}".format(ff))
                    sys.exit(0)

                hdulist = fits.open(ff)
                cube = hdulist[0].data
                cube_name = os.path.splitext(os.path.basename(ff))[0]
                header = hdulist[0].header
                # start drawing subprocess
                p = Process(target=draw_cube, args=(cube, cube_name, header, seeing, airmass, i))
                p.start()

                # printcube information

                print("\n{0}/{1} files".format(i+1, len(fitsfiles)))
                print("Cube: {0}".format(cube_name))

                # ask if cube is good or not
                keep_cube = None

                while keep_cube not in ['y', 'n', 'r']:
                    try:
                        print("\ty: keep")
                        print("\tn: discard")
                        print("\tr: flag for re-extraction")
                        keep_cube = input("\tChoose y/n/r: ").lower()[0]
                    except IndexError:
                        continue
                good_cubes[ff] = keep_cube
                hdulist.close()
                # close drawing subprocess
                p.terminate()
                p.join()
#            proc_cube_stats.terminate()
#            proc_cube_stats.join()
            plt.close('all')
            repeat = input("Finished viewing cubes. Return list and quit? (n to loop again) Y/n: ").lower()[0]
            if repeat == 'y':
                repeat = False
            else:
                continue
    # remove the 're-extraction' cubes
    reextract_list = []
    # convert good_cubes dict to Boolean
    """
    for key, val in good_cubes.iteritems():
        if val == 'y': 
            good_cubes[key] = True
        elif val == 'n': 
            good_cubes[key] = False
        elif val == 'r':
            reextract_list.append(key)
        else:
            good_cubes[key] = None
    """
    reextract_cubes = sorted([os.path.abspath(key) for key, val in good_cubes.iteritems() if val == 'r'])
    final_good_cubes = sorted([os.path.abspath(key) for key, val in good_cubes.iteritems() if val == 'y'])
    tot_exp_time = get_total_exposure_time(final_good_cubes, units.minute)
    if reextract_cubes:
        print("\n")
        print("Cubes flagged for re-extraction: ")
        for i in sorted(reextract_cubes): print(i)
    if final_good_cubes:
        print("\n")
        print("Combined exposure time for good cubes: {0:0.2f}".format(tot_exp_time))
    else:
        print("No good cubes")
    return final_good_cubes


#########################
# Spot checker ##########
#########################
def draw_spot_cube(cube, cube_name, spots):
    """
    Make a figure and draw cube slices on it
    spots are a list of [row, col] positions for each spot
    """
    # mask center for better image scaling
    cube[:,100:150,100:150] = np.nan #cube[:,100:150,100:150]*1e-3
    chan=0
    nchan = cube.shape[0]
    # get star positions
    #star_positions = P1640spots.get_single_cube_star_positions(np.array(spots))
    
    #try:
    fig = plt.figure()
    fig.suptitle(cube_name, fontsize='x-large')
    gridsize = (8,8)

    ax = fig.add_subplot(111)
    while True:
        ax.clear()
    
        chan = chan % nchan

        patches1 = [CirclePolygon(xy=spot[chan][::-1], radius=5,
                                  fill=False, alpha=1, ec='k', lw=2)
                    for spot in spots] # large circles centered on spot
        patches2 = [CirclePolygon(xy=spot[chan][::-1], radius=1,
                                  fill=True, alpha=0.3, ec='k', lw=2)
                    for spot in spots] # dots in location of spot
        #starpatches = [CirclePolygon(xy=star_positions[chan][::-1], radius=3,
        #                             fill=True, alpha=0.3, ec='k', lw=2)
        #               for spot in spots] # star position
        patchcoll = PatchCollection(patches1+patches2, match_original=True)
        
        imax = ax.imshow(cube[chan], norm=LogNorm())
        imax.axes.add_collection(patchcoll)
        ax.set_title("Channel {1:02d}".format(cube_name, chan))
        
        plt.pause(0.2)
        chan += 1
    #except KeyboardInterrupt:
    #    pass


# ALTERNATIVE TO REDEFINING SPOT CHECKER BUT THEN IT REQUIRES THE P1640_SPOT_CHECKER TO EXIST
#run_spot_checker = P1640_spot_checker.run_checker

def run_spot_checker(files=None, config=None, spot_path=None):
    """
    Supply ONE OF:
    files: list of files
    config: config file with a list of files
    """
    if spot_path is None:
        # set this default here so it can be overridden by a config
        configparser = ConfigParser.ConfigParser()
        configparser.read("../../P1640.ini")
        spot_directory = configparser.get("spots","spot_file_path")

    # files vs config: two args enter! one arg leaves!
    if not (files is None) != (config is None):
        print("Please supply either a list if files or a config file")
        return None
    elif files is not None:
        fitsfiles = files
    elif config is not None:
        configparser = ConfigParser.ConfigParser()
        configparser.read(config)
        # get the list of files
        fitsfiles = configparser.get("Input","occulted_files").split()
        spot_directory = configparser.get("Spots","spot_file_path")   
    else:
        print("Please supply either list of files or a config file")
        return None

    # spot path - if None, read default
    
    good_cubes = dict(zip(fitsfiles, [None for f in fitsfiles]))
    
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        #fig = plt.figure()
        for i, ff in enumerate(fitsfiles):
            # check file
            if not os.path.isfile(ff):
                print("File not found: {0}".format(ff))
                sys.exit(0)
                
            # get spots
            cubefile_name = os.path.splitext(os.path.basename(ff))[0]
            spot_files = glob.glob(os.path.join(spot_path, cubefile_name)+"*")
            if len(spot_files) == 0:
                print("No spot files found for {0}".format(os.path.basename(ff)))
                sys.exit(0)
            spots = [np.genfromtxt(f, delimiter=',') for f in spot_files]

            hdulist = fits.open(ff)
            cube = hdulist[0].data
            cube_name = os.path.splitext(os.path.basename(ff))[0]

            # start drawing subprocess
            p = Process(target=draw_spot_cube, args=(cube, cube_name, spots))
            p.start()

            # printcube information
            print("\n{0}/{1} files".format(i+1, len(fitsfiles)))
            print("\nCube: {0}".format(cube_name))
            print("\tExposure time: {0}".format(fits.getval(ff, "EXP_TIME")))
            print("\tSeeing: {0}".format(fits.getval(ff, "SEEING")))
            print("\tAirmass: {0}".format(np.mean([fits.getval(ff, "INIT_AM"),
                                                   fits.getval(ff, "FINL_AM")])))
            # ask if cube is good or not
            keep_cube = None
            while keep_cube not in ['y', 'n']:
                try:
                    keep_cube = input('\t\tKeep? y/n: ').lower()[0]
                except IndexError:
                    continue
            good_cubes[ff] = keep_cube

            # close drawing subprocess
            p.terminate()
            p.join()

        plt.close('all')
    for key, val in good_cubes.iteritems():
        if val == 'y': 
            good_cubes[key] = True
        elif val == 'n': 
            good_cubes[key] = False
        else:
            good_cubes[key] = None
    
    #print(good_cubes)
    #print("Good cubes: ")
    #for i in sorted([key for key, val inre good_cubes.iteritems() if val == True]):
    #    print(i)
    if np.all(good_cubes.values()):
        print("\nSpot fitting succeeded for all cubes.\n")
    else:
        print("\nSpot fitting failed for the following cubes:")
        print("\n".join([i for i in sorted(good_cubes.keys()) if good_cubes[i] == False]))
        print("\n")

    final_good_cubes = sorted([os.path.abspath(i) for i in good_cubes.keys() if good_cubes[i] == True])
    tot_exp_time = get_total_exposure_time(final_good_cubes, units.minute)
    print("Combined exposure time for good cubes: {0:0.2f} min".format(tot_exp_time))
    return final_good_cubes



def usage():
    print("""Required packages:
sys, os
warnings, multiprocessing
numpy, matplotlib
astropy
Usage:
    python cube_checker.py /path/to/data/cube.fits
    OR
    python cube_checker.py space.fits separated.fits sets.fits of.fits paths.fits to.fits cubes.fits
    OR
    python cube_checker.py `ls path/to/cubes/*fits`
    If running from IPython:
    files = glob.glob("all/the/fits/*fits")
    %run cube_checker.py {' '.join(files)}

Prints a list of the chosen file paths
""")

# Command-line option handling
# if switch --config is used, treat the following argument like the path to a config file
# if --config is not present, treat the following arguments like individual datacubes
class ConfigAction(argparse.Action):
    """
    Create a custom action to parse the 
    """
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        #if nargs is not None:
        #    raise ValueError("nargs not allowed")
        super(ConfigAction, self).__init__(option_strings, dest, **kwargs)
    def __call__(self, parser, namespace, values, option_string=None):
        configparser = ConfigParser.ConfigParser()
        configparser.read(values)
        # get the list of files
        filelist = configparser.get("Input","occulted_files").split()
        setattr(namespace, self.dest, filelist)
        try:
            setattr(namespace, 'spot_path', configparser.get("Spots","spot_file_path"))
        except:
            setattr(namespace, 'spot_path', None)


        
parser = argparse.ArgumentParser(prog="P1640_cube_checker.py",
                                 description='A utility to visually inspect P1640 cubes and spots')
group = parser.add_mutually_exclusive_group(required=True)
group.add_argument('--files', dest='files', nargs='*',
                    help='list of datacube files')
group.add_argument('--config', dest='files', nargs=1, action=ConfigAction,
                    help='config file containing fits files')
spotargs = parser.add_argument_group("spots", description="Use this if you want to check spot locations")
spotargs.add_argument("--spots", dest='spots',action='store_true', default=False,
                      help="use this flag if you want to overplot spot positions")
spotargs.add_argument("--spot_path", dest='spot_path', action='store',
                       default=dnah_spot_directory,
                       help='directory where spot position files are stored')
        
if __name__ == "__main__":

    if sys.argv[1] == 'help':
        usage()
        sys.exit()

    parseobj = parser.parse_args(sys.argv[1:])
    fitsfiles = parseobj.files

    if parseobj.spots:
        spot_directory = parseobj.spot_path
        good_cubes = run_spot_checker(fitsfiles, spot_path=spot_directory)
        #good_cubes = P1640_spot_checker.run_checker(fitsfiles, spot_path=spot_directory)
    else:
        good_cubes = run_checker(fitsfiles)
    print("Good cubes:")
    for i in good_cubes: print(i)




__author__ = 'JB'

from scipy.signal import convolve2d
from pyklip.kpp.stat.stat_utils import *
from pyklip.kpp.utils.oi import *


def point_source_detection(image, center,threshold,pix2as=None,mask_radius = 4,maskout_edge=None,maskout_inner_edge=None,IWA=None, OWA=None):
        """
        Find the brightest blobs in the image/cube.

        Args:
            image: Image from which to get the SNR map
            center: center of the image (y_cen, x_cen)
            threshold: Threshold under which blob should be ignore.
            pix2as: Platescale (arcsec per pixel).
            mask_radius: Radius of the mask used for masking point sources or the surroundings of the current pixel out
                        of the data. Default value is 4 pixels.
            maskout_edge: mask a maskout_edge (int) pixels around the outer edge of the FOV containing NaNs.
            maskout_inner_edge: mask a maskout_inner_edge (int) pixels border around the center nan disk.
            IWA: inner working angle in pixels.
            OWA: outer working angle in pixels.

        :return: Detection table..

            Table containing the list of the local maxima with their info
            Description by column: ["index","value","PA","Sep (pix)","Sep (as)","x","y","row","col"]
            1/ index of the candidate
            2/ Value of the maximum
            3/ Position angle in degree from North in [0,360]
            4/ Separation in pixel
            5/ Separation in arcsec
            6/ x position in pixel
            7/ y position in pixel
            8/ row index
            9/ column index
        """

        if image is not None:
            image = image
            if np.size(image.shape) == 2:
                ny,nx = image.shape
        if center is not None:
            center = [center]

        # Make a copy of the criterion map because it will be modified in the following.
        # Local maxima are indeed masked out when checked
        image_cpy = copy(image)
        stamp_size = mask_radius * 2 + 2
        image_cpy = np.pad(image_cpy,((stamp_size//2,stamp_size//2),(stamp_size//2,stamp_size//2)),mode="constant",constant_values=np.nan)

        # Build as grids of x,y coordinates.
        # The center is in the middle of the array and the unit is the pixel.
        # If the size of the array is even 2n x 2n the center coordinate in the array is [n,n].
        x_grid, y_grid = np.meshgrid(np.arange(0,nx,1)-center[0][0],np.arange(0,ny,1)-center[0][1])


        # Definition of the different masks used in the following.
        # Mask to remove the spots already checked in criterion_map.
        stamp_x_grid, stamp_y_grid = np.meshgrid(np.arange(0,stamp_size,1)-stamp_size//2,np.arange(0,stamp_size,1)-stamp_size//2)
        stamp_mask = np.ones((stamp_size,stamp_size))
        r_stamp = abs((stamp_x_grid) +(stamp_y_grid)*1j)
        stamp_mask[np.where(r_stamp < mask_radius)] = np.nan

        # Mask out a band of maskout_edge pixels around the edges of the finite pixels of the image.
        if maskout_edge is not None:
            IWA,OWA,inner_mask,outer_mask = get_occ(image_cpy, centroid = (center[0][0]+stamp_size//2,center[0][1]+stamp_size//2))
            conv_kernel = np.ones((maskout_edge,maskout_edge))
            flat_cube_wider_mask = convolve2d(outer_mask,conv_kernel,mode="same")
            image_cpy[np.where(np.isnan(flat_cube_wider_mask))] = np.nan
        if maskout_inner_edge is not None:
            IWA,OWA,inner_mask,outer_mask = get_occ(image_cpy, centroid = (center[0][0]+stamp_size//2,center[0][1]+stamp_size//2))
            conv_kernel = np.ones((maskout_inner_edge,maskout_inner_edge))
            flat_cube_wider_mask = convolve2d(inner_mask,conv_kernel,mode="same")
            image_cpy[np.where(np.isnan(flat_cube_wider_mask))] = np.nan


        # Number of rows and columns to add around a given pixel in order to extract a stamp.
        row_m = int(np.floor(stamp_size/2.0))    # row_minus
        row_p = int(np.ceil(stamp_size/2.0))     # row_plus
        col_m = int(np.floor(stamp_size/2.0))    # col_minus
        col_p = int(np.ceil(stamp_size/2.0))     # col_plus

        # Table containing the list of the local maxima with their info
        # Description by column:
        # 1/ index of the candidate
        # 2/ Value of the maximum
        # 3/ Position angle in degree from North in [0,360]
        # 4/ Separation in pixel
        # 5/ Separation in arcsec
        # 6/ x position in pixel
        # 7/ y position in pixel
        # 8/ row index
        # 9/ column index
        candidates_table = []
        table_labels = ["index","value","PA","Sep (pix)","Sep (as)","x","y","row","col"]
        ## START WHILE LOOP.
        # Each iteration looks at one local maximum in the criterion map.
        k = 0
        max_val_criter = np.nanmax(image_cpy)
        while max_val_criter > threshold:# and k <= max_attempts:
            k += 1
            # Find the maximum value in the current criterion map. At each iteration the previous maximum is masked out.
            max_val_criter = np.nanmax(image_cpy)
            # Locate the maximum by retrieving its coordinates
            max_ind = np.where( image_cpy == max_val_criter )
            row_id,col_id = max_ind[0][0],max_ind[1][0]
            x_max_pos = x_grid[row_id-stamp_size//2,col_id-stamp_size//2]
            y_max_pos = y_grid[row_id-stamp_size//2,col_id-stamp_size//2]
            sep_pix = np.sqrt(x_max_pos**2+y_max_pos**2)
            if pix2as is not None:
                sep_arcsec = pix2as *sep_pix
            else:
                sep_arcsec = 0
            pa = np.mod(np.rad2deg(np.arctan2(-x_max_pos,y_max_pos)),360)

            # Mask the spot around the maximum we just found.
            image_cpy[(row_id-row_m):(row_id+row_p), (col_id-col_m):(col_id+col_p)] *= stamp_mask

            if IWA is not None:
                if sep_pix < IWA:
                    continue
            if OWA is not None:
                if sep_pix > OWA:
                    continue

            # Store the current local maximum information in the table
            candidates_table.append([k,max_val_criter,pa,sep_pix,sep_arcsec,x_max_pos,y_max_pos,row_id-stamp_size//2,col_id-stamp_size//2])
        ## END WHILE LOOP.

        return candidates_table


import numpy as np
from scipy.interpolate import interp2d

def samplingRegion(size_window, theta = [45, 135], m = 0.2, M = 0.8, step = 1, decimals = 2, ray = False):
    """This function returns all the coordinates of the sampling region, the center of the region is (0,0)
    When applying to matrices, don't forget to SHIFT THE CENTER!
    Input:
        size_window: the radius of the sampling region. The whole region should thus have a length of 2*size_window+1.
        theta: the angle range of the sampling region, default: [45, 135] for the anti-diagonal and diagonal directions.
        m: the minimum fraction of size_window, default: 0.2 (i.e., 20%). In this way, the saturated region can be excluded.
        M: the maximum fraction of size_window, default: 0.8 (i.e., 80%). Just in case if there's some star along the diagonals.
        step: the seperation between sampling dots (units: pixel), default value is 1pix.
        decimals: the precisoin of the sampling dots (units: pixel), default value is 0.01pix.
        ray: only half of the line?
    Output: (xs, ys)
        xs: x indecies, flattend.
        ys: y indecies, flattend.
    Example:
        1. If you call "xs, ys = samplingRegion(5)", you will get:
        xs: array([-2.83, -2.12, -1.41, -0.71,  0.71,  1.41,  2.12,  2.83,  2.83, 2.12,  1.41,  0.71, -0.71, -1.41, -2.12, -2.83]
        ys: array([-2.83, -2.12, -1.41, -0.71,  0.71,  1.41,  2.12,  2.83, -2.83, -2.12, -1.41, -0.71,  0.71,  1.41,  2.12,  2.83]))
        2. For "radonCenter.samplingRegion(5, ray=True)", you will get:
        xs: array([ 0.71,  1.41,  2.12,  2.83, -0.71, -1.41, -2.12, -2.83])
        ys: array([ 0.71,  1.41,  2.12,  2.83,  0.71,  1.41,  2.12,  2.83])
    """
    
    if np.asarray(theta).shape == ():
        theta = [theta]
    #When there is only one angle
        
    theta = np.array(theta)
    if ray:
        zeroDegXs = np.arange(int(size_window*m), int(size_window*M) + 0.1 * step, step)
    else:
        zeroDegXs = np.append(np.arange(-int(size_window*M), -int(size_window*m) + 0.1 * step, step), np.arange(int(size_window*m), int(size_window*M) + 0.1 * step, step))
    #create the column indecies if theta = 0
    zeroDegYs = np.zeros(zeroDegXs.size)
    
    xs = np.zeros((np.size(theta), np.size(zeroDegXs)))
    ys = np.zeros((np.size(theta), np.size(zeroDegXs)))
    
    for i, angle in enumerate(theta):
        degRad = np.deg2rad(angle)
        angleDegXs = np.round(zeroDegXs * np.cos(degRad), decimals = decimals)
        angleDegYs = np.round(zeroDegXs * np.sin(degRad), decimals = decimals)
        xs[i, ] = angleDegXs
        ys[i, ] = angleDegYs
    
    xs = xs.flatten()
    ys = ys.flatten()

    return xs, ys


def smoothCostFunction(costFunction, halfWidth = 0):
    """
    smoothCostFunction will smooth the function within +/- halfWidth, i.e., to replace the value with the average within +/- halfWidth pixel.
    This function can be genrally used to smooth any 2D matrix.
    Input:
        costFunction: original cost function, a matrix.
        halfWdith: the half width of the smoothing region, default = 0 pix.
    Output:
        newFunction: smoothed cost function.
    """
    if halfWidth == 0:
        return costFunction
    else:
        newFunction = np.zeros(costFunction.shape)
        rowRange = np.arange(costFunction.shape[0], dtype=int)
        colRange = np.arange(costFunction.shape[1], dtype=int)
        rangeShift = np.arange(-halfWidth, halfWidth + 0.1, dtype=int)
        for i in rowRange:
            for j in colRange:
                if np.isnan(costFunction[i, j]):
                    newFunction[i, j] = np.nan
                else:
                    surrondingNumber = (2 * halfWidth + 1) ** 2
                    avg = 0
                    for ii in (i + rangeShift):
                        for jj in (j + rangeShift):
                            if (not (ii in rowRange)) or (not (jj in colRange)) or (np.isnan(costFunction[ii, jj])):
                                surrondingNumber -= 1
                            else:
                                avg += costFunction[ii, jj]
                    newFunction[i, j] = avg * 1.0 / surrondingNumber
    return newFunction
    
 
def searchCenter(image, x_ctr_assign, y_ctr_assign, size_window, m = 0.2, M = 0.8, size_cost = 5, theta = [45, 135], ray = False, smooth = 2, decimals = 2, output_cost=False):
    """
    This function searches the center in a grid, 
    calculate the cost function of Radon Transform (Pueyo et al., 2015), 
    then interpolate the cost function, 
    get the center which corresponds to the maximum value in the cost function.
    
    Input:
        image: 2d array.
        x_ctr_assign: the assigned x-center, or starting x-position; for STIS, the "CRPIX1" header is suggested.
        x_ctr_assign: the assigned y-center, or starting y-position; for STIS, the "CRPIX2" header is suggested.
        size_window: half width of the sampling region; size_window = image.shape[0]/2 is suggested.
            m & M:  The sampling region will be (-M*size_window, -m*size_window)U(m*size_window, M*size_window).
        size_cost: search the center within +/- size_cost pixels, i.e., a square region.
        theta: the angle range of the sampling region; default: [45, 135] for the anti-diagonal and diagonal directions.
        ray: is the theta a line or a ray? Default: line.
        smooth: smooth the cost function, for one pixel, replace it by the average of its +/- smooth neighbours; defualt = 2.
        decimals: the precision of the centers; default = 2 for a precision of 0.01.
    Output:
        x_cen, y_cen
    """
    (y_len, x_len) = image.shape

    x_range = np.arange(x_len)
    y_range = np.arange(y_len)

    image_interp = interp2d(x_range, y_range, image, kind = 'cubic')
    #interpolate the image
    
    
    precision = 1
    x_centers = np.round(np.arange(x_ctr_assign - size_cost, x_ctr_assign + size_cost + precision/10.0, precision), decimals=1)
    y_centers = np.round(np.arange(y_ctr_assign - size_cost, y_ctr_assign + size_cost + precision/10.0, precision), decimals=1)
    costFunction = np.zeros((x_centers.shape[0], y_centers.shape[0]))
    #The above 3 lines create the centers of the search region
    #The cost function stores the sum of all the values in the sampling region
    
    size_window = size_window - size_cost
    (xs, ys) = samplingRegion(size_window, theta, m = m, M = M, ray = ray)
    #the center of the sampling region is (0,0), don't forget to shift the center!
    
    for j, x0 in enumerate(x_centers):
        for i, y0 in enumerate(y_centers):
            value = 0
            
            for x1, y1 in zip(xs, ys):
                #Shifting the center, this now is the coordinate of the RAW IMAGE
                x = x0 + x1
                y = y0 + y1
                
                value += image_interp(x, y)
            
            costFunction[i, j] = value  #Create the cost function

    costFunction = smoothCostFunction(costFunction, halfWidth = smooth)
    #Smooth the cost function
    
    interp_costfunction = interp2d(x_centers, y_centers, costFunction, kind='cubic')
    
    
    for decimal in range(1, decimals+1):
        precision = 10**(-decimal)
        if decimal >= 2:
            size_cost = 10*precision
        x_centers_new = np.round(np.arange(x_ctr_assign - size_cost, x_ctr_assign + size_cost + precision/10.0, precision), decimals=decimal)
        y_centers_new = np.round(np.arange(y_ctr_assign - size_cost, y_ctr_assign + size_cost + precision/10.0, precision), decimals=decimal)
    
        x_cen = 0
        y_cen = 0
        maxcostfunction = 0
        value = np.zeros((y_centers_new.shape[0], x_centers_new.shape[0]))
    
        for j, x in enumerate(x_centers_new):
            for i, y in enumerate(y_centers_new):
                value[i, j] = interp_costfunction(x, y)
        
        idx = np.where(value == np.max(value))
        #Just in case when there are multile maxima, then use the average of them. 
        x_cen = np.mean(x_centers_new[idx[1]])
        y_cen = np.mean(y_centers_new[idx[0]])
        
        x_ctr_assign = x_cen
        y_ctr_assign = y_cen
    
    x_cen = round(x_cen, decimals)
    y_cen = round(y_cen, decimals)

    if output_cost:
        return x_cen, y_cen, costFunction
    else:
        return x_cen, y_cen



# pylint: disable=C0103
from sys import version_info
from os import path
import multiprocessing as mp
from copy import deepcopy
import ctypes

import pickle
import h5py

import numpy as np
import scipy.ndimage as ndimage

from pyklip.fmlib.nofm import NoFM
import pyklip.fm as fm
from pyklip.klip import rotate

# define the global variables for that code
class DiskFM(NoFM):
    """Defining a model disk to which we apply the Forward Modelling. There are 3 ways:

            * "Save Basis mode" (save_basis=true), we are preparing to save the FM basis
            * "Load Basis mode" (load_from_basis = true), most of the parameters are
              derived from the previous fm.klip_dataset which measured FM basis.
            * "Simple FM mode" (save_basis = load_from_basis = False). Just
              for a unique disk FM.


        Args:
            inputs_shape: shape of the inputs numpy array. Typically (N, x, y)
            numbasis: 1d numpy array consisting of the number of basis vectors to use
            dataset: an instance of Instrument.Data. We need it to know the
                     parameters to "prepare" first inital model.
            model_disk: a model of the disk of size (wvs, x, y) or (x, y)
            basis_filename: filename to save and load the KL basis. Filenames can haves
                            2 recognizable extensions: .h5 or .pkl. We strongly
                            recommand .h5 as pickle have problem of compatibility
                            between python 2 and 3 and sometimes between computer
                            (e.g. KL modes not readable on another computer)
            load_from_basis: if True, load the KL basis at basis_filename. It only need
                             to be done once, after which you can measure FM with
                             only update_model()
            save_basis: if True, save the KL basis at basis_filename. If load_from_basis
                            is True, save_basis is automatically set to False, it is
                            useless to load and save the matrix at the same time.
            aligned_center: array of 2 elements [x,y] that all the model will be
                            centered on for image registration.
                            FIXME: This is the most problematic thing currently, the
                            aligned_center of the model and of the images can be set
                            independently, which will create false results.
                            - In "Load Basis mode", this parameter is not read, we just
                            use the aligned_center set for the images in the previous
                            fm.klip_dataset and save in basis_filename
                            - In "Save Basis mode", or "Simple FM mode" we define it
                            and then check that it is the same one used for the images
                            in fm.klip_dataset
            mode: deprecated parameter, ignored here and defined in fm.klip_dataset
            annuli: deprecated parameter, ignored here and defined in fm.klip_dataset
            subsections: deprecated parameter, ignored here and defined
                         in fm.klip_dataset
            numthreads: deprecated parameter. All centering are done in fm.klip_dataset 

        Returns:
            A DiskFM Object

    """
    def __init__(self,
                 inputs_shape,
                 numbasis,
                 dataset,
                 model_disk,
                 basis_filename="",
                 kl_basis_file=None,
                 load_from_basis=False,
                 save_basis=False,
                 aligned_center=None,
                 psf_library=None,
                 mode=None,
                 annuli=None,
                 subsections=None,
                 numthreads=None):
        """

            Initilaizes the DiskFM class

        """

        if load_from_basis:
            numbasis = 1
            inputs_shape = 1
            dataset = None

        # make sure the dimensions have the good shape
        # and that they are numpy arrays to access their shape
        if hasattr(numbasis, "__len__"):
            numbasis = np.array(numbasis)
        else:
            numbasis = np.array([numbasis])

        if hasattr(inputs_shape, "__len__"):
            inputs_shape = np.array(inputs_shape)
        else:
            inputs_shape = np.array([inputs_shape])

        if mode is not None:
            print(mode)
            print("Warning: Argument 'mode' in pyklip.fmlib.diskfm.DiskFM class definition "\
                "is deprecated (not used and will be removed in a future version). " \
                "KLIP reduction parameters (mode, annuli and subsections) "\
                "are only defined once in klip_dataset [pyklip.parallelized]. ")

        if numthreads is not None:
            print("Warning: Argument 'numthreads' in pyklip.fmlib.diskfm.DiskFM class definition "\
                "is deprecated (not used and will be removed in a future version). " \
                "All centering are done in fm.klip_dataset ")

        if annuli is not None:
            print("Warning: Argument 'annuli' in pyklip.fmlib.diskfm.DiskFM class definition "\
                "is deprecated (not used and will be removed in a future version). " \
                "KLIP reduction parameters (mode, annuli and subsections) "\
                "are only defined once in klip_dataset [pyklip.parallelized].")

        if subsections is not None:
            print("Warning: Argument 'subsections' in pyklip.fmlib.diskfm.DiskFM class definition "\
                "is deprecated (not used and will be removed in a future version). " \
                "KLIP reduction parameters (mode, annuli and subsections) "\
                "are only defined once in klip_dataset [pyklip.parallelized].")

        super(DiskFM, self).__init__(inputs_shape, numbasis)

        self.data_type = ctypes.c_double

        self.basis_filename = basis_filename
        self.save_basis = save_basis
        self.load_from_basis = load_from_basis

        if self.load_from_basis:
            # Its useless to save and load at the same time.
            self.save_basis = False
            save_basis = False

        if self.save_basis is True:
            manager = mp.Manager()
            self.klmodes_dict = manager.dict()
            self.evecs_dict = manager.dict()
            self.evals_dict = manager.dict()
            self.aligned_images_dict = manager.dict()
            self.ref_psfs_indicies_dict = manager.dict()
            self.section_ind_dict = manager.dict()

            self.radstart_dict = manager.dict()
            self.radend_dict = manager.dict()
            self.phistart_dict = manager.dict()
            self.phiend_dict = manager.dict()
            self.input_img_num_dict = manager.dict()

            self.klparam_dict = manager.dict()
        # Coords where align_and_scale places model center

        if self.load_from_basis is True:  # We want to load the FM basis

            # We load the FM basis files, before preparing the model to
            # be sure that the aligned_center is identical to the one used
            # when measuring the KL
            self.load_basis_files(psf_library=psf_library,
                                  kl_basis_file=kl_basis_file)

        else:  # We want to save the basis or just a single disk FM

            # Attributes of input
            self.inputs_shape = inputs_shape

            self.numbasis = numbasis

            # Outputs attributes
            output_imgs_shape = inputs_shape + self.numbasis.shape
            self.output_imgs_shape = output_imgs_shape

            self.PAs = dataset.PAs
            self.wvs = dataset.wvs

            self.nwvs = int(np.size(np.unique(
                self.wvs)))  # Get the number of wvls
            self.nfiles = int(self.inputs_shape[0] /
                              self.nwvs)  # Get the number of files

            # default aligned_center if none (same default as fm.parallelized):
            if aligned_center is None:
                centers = dataset.centers
                aligned_center = [
                    np.mean(centers[:, 0]),
                    np.mean(centers[:, 1])
                ]

            # define the center
            self.aligned_center = aligned_center

        # Prepare the first disk for FM
        self.update_disk(model_disk)

    def update_disk(self, model_disk):
        """
        Takes model disk and rotates it to the PAs of the input images for use as
        reference PSFS

        The disk can be either an 3D array of shape (wvs,y,x) for data of the same shape
        or a 2D Array of shape (y,x), in which case, if the dataset is multiwavelength
        the same model is used for all wavelenths.

        Args:
            model_disk: Disk to be forward modeled.

        Returns:
            None
        """

        self.model_disks = np.zeros(self.inputs_shape)

        # Extract the # of WL per files
        n_wv_per_file = self.nwvs  # Number of wavelenths per file.

        model_disk_shape = np.shape(model_disk)

        if (np.size(model_disk_shape) == 2) & (n_wv_per_file > 1):
            # This is a single WL 2D model in a multi-wl 3D data,
            # in that case we repeat this model at each WL
            self.model_disk = np.broadcast_to(model_disk, (n_wv_per_file, ) +
                                              model_disk.shape)
            model_disk_shape = np.shape(model_disk)
        else:
            # This is either a multi WL 3D model in a multi-wl 3D data
            # or a single WL 3D model in a single-wl 2D data, we do nothing
            self.model_disk = model_disk

        # Check if we have a disk at multiple wavelengths
        if np.size(model_disk_shape) > 2:  # Then it's a multiWL model
            n_model_wvs = model_disk_shape[0]

            if n_model_wvs != n_wv_per_file:
                # Both models and data are multiWL, but not the same number of WLs !
                raise ValueError(
                    """Number of wls in disk model ({0}) don't match number of wls in
                    the data ({1})""".format(n_model_wvs, n_wv_per_file))

            for k in np.arange(self.nfiles):
                for j, _ in enumerate(range(n_model_wvs)):
                    model_copy = deepcopy(model_disk[j, :, :])
                    model_copy = rotate(
                        model_copy,
                        self.PAs[k * n_wv_per_file + j],
                        self.aligned_center,
                        flipx=True,
                    )
                    model_copy[np.where(np.isnan(model_copy))] = 0.0
                    self.model_disks[k * n_wv_per_file + j, :, :] = model_copy

        else:  # This is a 2D disk model and a wl = 1 case

            for i, pa_here in enumerate(self.PAs):
                model_copy = deepcopy(model_disk)
                model_copy = rotate(model_copy,
                                    pa_here,
                                    self.aligned_center,
                                    flipx=True)
                model_copy[np.where(np.isnan(model_copy))] = 0.0
                self.model_disks[i] = model_copy

        self.model_disks = np.reshape(
            self.model_disks,
            (self.inputs_shape[0],
             self.inputs_shape[1] * self.inputs_shape[2]),
        )

    def alloc_fmout(self, output_img_shape):
        """Allocates shared memory for the output of the shared memory


        Args:
            output_img_shape: shape of output image (usually N,y,x,b)

        Returns:
            [mp.array to store FM data in, shape of FM data array]

        """

        fmout_size = int(np.prod(output_img_shape))
        fmout_shape = output_img_shape
        fmout = mp.Array(self.data_type, fmout_size)
        return fmout, fmout_shape

    def fm_from_eigen(self,
                      klmodes=None,
                      evals=None,
                      evecs=None,
                      input_img_shape=None,
                      output_img_shape=None,
                      input_img_num=None,
                      ref_psfs_indicies=None,
                      section_ind=None,
                      aligned_imgs=None,
                      radstart=None,
                      radend=None,
                      phistart=None,
                      phiend=None,
                      padding=None,
                      IOWA=None,
                      ref_center=None,
                      parang=None,
                      numbasis=None,
                      fmout=None,
                      flipx=True,
                      mode=None,
                      **kwargs):
        """
        Generate forward models using the KL modes, eigenvectors, and eigenvectors from
        KLIP. Calls fm.py functions to perform the forward modelling. If we wish to save
        the KL modes, it save in dictionnaries.

        Args:
            klmodes: unpertrubed KL modes
            evals: eigenvalues of the covariance matrix that generated the KL modes in
                    ascending order(lambda_0 is the 0 index) (shape of [nummaxKL])
            evecs: corresponding eigenvectors (shape of [p, nummaxKL])
            input_image_shape: 2-D shape of inpt images ([ysize, xsize])
            input_img_num: index of sciece frame
            ref_psfs_indicies: array of indicies for each reference PSF
            section_ind: array indicies into the 2-D x-y image that correspond to
                            this section. Note: needs be called as section_ind[0]
            radstart: radius of start of segment
            radend: radius of end of segment
            phistart: azimuthal start of segment [radians]
            phiend: azimuthal end of segment [radians]
            padding: amount of padding on each side of sector
            IOWA: tuple (IWA,OWA) IWA = Inner working angle & OWA = Outer working angle,
                    both in pixels. It defines the separation interva in which klip will
                    be run.
            ref_center: center of image
            parang: parallactic angle of input image [DEGREES]
            numbasis: array of KL basis cutoffs
            fmout: numpy output array for FM output. Shape is (N, y, x, b)
            mode: mode of the reduction ('RDI', 'ADI', 'SDI'). If RDI only, we only 
                    measure the oversubctraction
            kwargs: any other variables that we don't use but are part of the input

        Returns:
            None

        """

        # we check that the aligned_center used to center the disk (self.aligned_center)
        # If the same used to center the image in klip_dataset.
        # If not, we should not continue.
        if self.aligned_center != ref_center:
            err_string = """The aligned_center for the model {0} and for
                            the data {1} is different.
                            Change and rerun""".format(self.aligned_center,
                                                       ref_center)

            print(err_string)
            raise Exception(err_string)

        if self.load_from_basis == False:
            sci = aligned_imgs[input_img_num, section_ind[0]]
            refs = aligned_imgs[ref_psfs_indicies, :]
            refs = refs[:, section_ind[0]]
        else:
            wlstrkey = 'wl' + str(int(self.wvs[input_img_num] * 1000)).zfill(4)
            sci = self.aligned_images_dict[wlstrkey][input_img_num,
                                                     section_ind[0]]
            # in the case of load_from_basis, the images are already
            # saved in the DiskFM object, we can save a few tens of
            # Mbytes (per cpu) by not saving them
            # and just passing them to the nex function

        # use the disk model stored
        model_sci = self.model_disks[input_img_num, section_ind[0]]
        model_sci[np.where(np.isnan(model_sci))] = 0
        model_ref = self.model_disks[ref_psfs_indicies, :]
        model_ref = model_ref[:, section_ind[0]]
        model_ref[np.where(np.isnan(model_ref))] = 0
        if mode == 'RDI':
            #if only RDI we skip the deltaKL calculation since we do only over-subctraction
            delta_KL = klmodes * 0.
        else:
            # using original Kl modes and reference models, compute the perturbed KL modes
            # (spectra is already in models)
            if self.load_from_basis == False:
                delta_KL = fm.perturb_specIncluded(
                    evals,
                    evecs,
                    klmodes,
                    refs,
                    model_ref,
                    return_perturb_covar=False,
                )
            else:
                # in the case of load_from_basis, the images are already saved in the
                # DiskFM object, we can save a few tens of Mbytes (per cpu) by not 
                # saving them and just passing them to the nex function
                delta_KL = fm.perturb_specIncluded(
                    evals,
                    evecs,
                    klmodes,
                    self.aligned_images_dict[wlstrkey][ref_psfs_indicies, :]
                    [:, section_ind[0]],
                    model_ref,
                    return_perturb_covar=False,
                )

        # calculate postklip_psf using delta_KL
        postklip_psf, _, _ = fm.calculate_fm(delta_KL,
                                             klmodes,
                                             numbasis,
                                             sci,
                                             model_sci,
                                             inputflux=None)

        # write forward modelled disk to fmout (as output)
        # need to derotate the image in this step

        for thisnumbasisindex in range(np.size(numbasis)):
            fm._save_rotated_section(input_img_shape,
                                     postklip_psf[thisnumbasisindex],
                                     section_ind,
                                     fmout[input_img_num, :, :,
                                           thisnumbasisindex],
                                     None,
                                     parang,
                                     radstart,
                                     radend,
                                     phistart,
                                     phiend,
                                     padding,
                                     IOWA,
                                     ref_center,
                                     flipx=flipx)

        # We save the KL basis and params for this image and section in a dictionnaries
        if self.save_basis is True:
            # save the parameter used in KLIP-FM. We save a float64 to avoid pbs
            # in the saving and loading

            if mode == 'RDI':
                self.klparam_dict['isRDI'] = np.float64(1.)
            else:
                self.klparam_dict['isRDI'] = np.float64(0.)

            [IWA, OWA] = IOWA
            self.klparam_dict['IWA'] = np.float64(IWA)
            self.klparam_dict['OWA'] = np.float64(OWA)

            self.klparam_dict['input_img_shape'] = np.float64(input_img_shape)
            self.klparam_dict['numbasis'] = np.float64(numbasis)
            self.klparam_dict['output_imgs_shape'] = np.float64(
                output_img_shape)

            # To have a single identifier for each set of aligned images,
            # we save the wavelenght in nm
            wlstrkey = 'wl' + str(int(self.wvs[input_img_num] * 1000)).zfill(4)
            self.aligned_images_dict[wlstrkey] = aligned_imgs

            # save the center for aligning the image in KLIP-FM. In practice, this
            # center will be used for all the models after we load.
            self.klparam_dict['aligned_center_x'] = np.float64(ref_center[0])
            self.klparam_dict['aligned_center_y'] = np.float64(ref_center[1])

            # We save information about the dataset that will be used when we load the KL basis
            self.klparam_dict['PAs'] = np.float64(self.PAs)
            self.klparam_dict['wvs'] = np.float64(self.wvs)

            self.klparam_dict['nwvs'] = np.float64(self.nwvs)
            self.klparam_dict['nfiles'] = np.float64(self.nfiles)

            # To have a single identifier for each set of section/image for the
            # dictionnaries key, we use section first pixel and image number
            curr_im = str(input_img_num).zfill(3)
            namkey = 'idsec' + str(section_ind[0][0]) + 'i' + curr_im
            # saving the KL modes dictionnaries
            self.klmodes_dict[namkey] = klmodes
            self.evals_dict[namkey] = evals
            self.evecs_dict[namkey] = evecs
            self.ref_psfs_indicies_dict[namkey] = ref_psfs_indicies
            self.section_ind_dict[namkey] = section_ind

            # saving the section delimiters dictionnaries
            self.radstart_dict[namkey] = radstart
            self.radend_dict[namkey] = radend
            self.phistart_dict[namkey] = phistart
            self.phiend_dict[namkey] = phiend
            self.input_img_num_dict[namkey] = input_img_num

    def cleanup_fmout(self, fmout):
        """
        After running KLIP-FM, we need to reshape fmout so that the numKL dimension is
        the first one and not the last. We also use this function to save the KL basis
        because it is called by fm.py at the end fm.klip_parallelized

        Args:
            fmout: numpy array of ouput of FM

        Returns:
            Same but cleaned up if necessary
        """

        # save the KL basis.
        if self.save_basis:
            self.save_kl_basis()

        # FIXME We save the matrix here it here because it is called by fm.py at the end
        # fm.klip_parallelized but this is not ideal.

        dims = fmout.shape
        fmout = np.rollaxis(
            fmout.reshape((dims[0], dims[1], dims[2], dims[3])), 3)
        return fmout

    def save_fmout(self,
                   dataset,
                   fmout,
                   outputdir,
                   fileprefix,
                   numbasis,
                   klipparams=None,
                   calibrate_flux=False,
                   pixel_weights=1,
                   **kwargs):
        """
        Uses dataset parameters to save the forward model, the output of
        fm_paralellized or klip_dataset. No returm, data are saved
        in "fileprefix" .fits files

        Args:
            dataset: an instance of Instrument.Data . Will use its
                     dataset.savedata() function to save data
            fmout: output of forward modelling.
            outputdir: directory to save output files
            fileprefix: filename prefix for saved files
            numbasis: number of KL basis vectors to use
                      (can be a scalar or list like)
            klipparams: string with KLIP-FM parameters
            calibrate_flux: if True, flux calibrate the data in the same way as
                            the klipped data
            pixel_weights: weights for each pixel for weighted mean. Leave this as a
                           single number for simple mean

        Returns:
            None

        """

        weighted = len(np.shape(pixel_weights)) > 1
        numwvs = dataset.numwvs
        fmout_spec = fmout.reshape([
            fmout.shape[0],
            fmout.shape[1] // numwvs,
            numwvs,
            fmout.shape[2],
            fmout.shape[3],
        ])  # (b, N_cube, wvs, y, x) 5-D cube

        # collapse in time and wavelength to examine KL modes
        KLmode_cube = np.nanmean(pixel_weights * fmout_spec, axis=(1, 2))
        if weighted:
            # if the pixel weights aren't just 1 (i.e., weighted case),
            # we need to normalize for that
            KLmode_cube /= np.nanmean(pixel_weights, axis=(1, 2))

        # broadband flux calibration for KL mode cube
        if calibrate_flux:
            KLmode_cube = dataset.calibrate_output(KLmode_cube, spectral=False)

        dataset.savedata(
            path.join(outputdir, fileprefix + "-fmpsf-KLmodes-all.fits"),
            KLmode_cube,
            klipparams=klipparams.format(numbasis=str(numbasis)),
            filetype="KL Mode Cube",
            zaxis=numbasis,
        )

        # if there is more than one wavelength, save also spectral cubes
        if dataset.numwvs > 1:

            KLmode_spectral_cubes = np.nanmean(pixel_weights * fmout_spec,
                                               axis=1)
            if weighted:
                # if the pixel weights aren't just 1 (i.e., weighted case), we need to
                # normalize for that.
                KLmode_spectral_cubes /= np.nanmean(pixel_weights, axis=1)

            for KLcutoff, spectral_cube in zip(numbasis,
                                               KLmode_spectral_cubes):
                # calibrate spectral cube if needed
                if calibrate_flux:
                    spectral_cube = dataset.calibrate_output(spectral_cube,
                                                             spectral=True)
                dataset.savedata(
                    path.join(
                        outputdir, fileprefix +
                        "-fmpsf-KL{0}-speccube.fits".format(KLcutoff)),
                    spectral_cube,
                    klipparams=klipparams.format(numbasis=KLcutoff),
                    filetype="PSF Subtracted Spectral Cube",
                )

    def save_kl_basis(self):
        """
        Save the KL basis and other needed parameters

        Args:
            None

        Returns:
            None

        """

        # Convert everything to np arrays and types to be safe for the saving.
        for key in self.section_ind_dict.keys():
            self.section_ind_dict[key] = np.asarray(self.section_ind_dict[key])
            self.radstart_dict[key] = np.float64(self.radstart_dict[key])
            self.radend_dict[key] = np.float64(self.radend_dict[key])
            self.phistart_dict[key] = np.float64(self.phistart_dict[key])
            self.phiend_dict[key] = np.float64(self.phiend_dict[key])

        _, file_extension = path.splitext(self.basis_filename)

        if file_extension == ".pkl":
            # transform mp dicts to normal dicts
            pkl_file = open(self.basis_filename, "wb")

            pickle.dump(dict(aligned_images_dict), pkl_file, protocol=2)

            pickle.dump(dict(klmodes_dict), pkl_file, protocol=2)
            pickle.dump(dict(evecs_dict), pkl_file, protocol=2)
            pickle.dump(dict(evals_dict), pkl_file, protocol=2)
            pickle.dump(dict(ref_psfs_indicies_dict), pkl_file, protocol=2)
            pickle.dump(dict(section_ind_dict), pkl_file, protocol=2)

            pickle.dump(dict(radstart_dict), pkl_file, protocol=2)
            pickle.dump(dict(radend_dict), pkl_file, protocol=2)
            pickle.dump(dict(phistart_dict), pkl_file, protocol=2)
            pickle.dump(dict(phiend_dict), pkl_file, protocol=2)
            pickle.dump(dict(input_img_num_dict), pkl_file, protocol=2)

            pickle.dump(dict(klparam_dict), pkl_file, protocol=2)

        elif file_extension == ".h5":
            # transform mp dicts to normal dicts
            # make a single dictionnary and save in h5

            saving_in_h5_dict = {
                'aligned_images_dict': dict(self.aligned_images_dict),
                'klmodes_dict': dict(self.klmodes_dict),
                'evecs_dict': dict(self.evecs_dict),
                'evals_dict': dict(self.evals_dict),
                'ref_psfs_indicies_dict': dict(self.ref_psfs_indicies_dict),
                'section_ind_dict': dict(self.section_ind_dict),
                'radstart_dict': dict(self.radstart_dict),
                'radend_dict': dict(self.radend_dict),
                'phistart_dict': dict(self.phistart_dict),
                'phiend_dict': dict(self.phiend_dict),
                'input_img_num_dict': dict(self.input_img_num_dict),
                'klparam_dict': dict(self.klparam_dict),
            }

            _save_dict_to_hdf5(saving_in_h5_dict, self.basis_filename)

            del saving_in_h5_dict

        else:
            raise ValueError(file_extension +
                             """ is not a possible extension. Filenames can
                haves 2 recognizable extension2: .h5 and .pkl""")

    def load_basis_files(self, psf_library=None, kl_basis_file=None):
        """
        Loads in previously saved basis files and sets variables for fm_from_eigen

        Args:
            dataset: an instance of Instrument.Data, after fm.klip_dataset.
                     Allow me to pass in the structure some correction parameters
                     set by fm.klip_dataset, such as IWA, OWA, aligned_center.
                     KL basis and sections information are passed via global variables

        Returns:
            None
        """
        if kl_basis_file is not None:
            file_extension = ""
        else:
            _, file_extension = path.splitext(self.basis_filename)
        manager = mp.Manager()

        # Load in file
        if file_extension == ".pkl":
            pkl_file = open(self.basis_filename, "rb")
            if version_info.major == 3:
                # Using encoding='latin1' is required for unpickling NumPy arrays
                # and instances of datetime, date and time pickled by Python 2.
                self.aligned_images_dict = manager.dict(
                    pickle.load(pkl_file, encoding="latin1"))

                self.klmodes_dict = manager.dict(
                    pickle.load(pkl_file, encoding="latin1"))
                self.evecs_dict = manager.dict(
                    pickle.load(pkl_file, encoding="latin1"))
                self.evals_dict = manager.dict(
                    pickle.load(pkl_file, encoding="latin1"))
                self.ref_psfs_indicies_dict = manager.dict(
                    pickle.load(pkl_file, encoding="latin1"))
                self.section_ind_dict = manager.dict(
                    pickle.load(pkl_file, encoding="latin1"))

                self.radstart_dict = manager.dict(
                    pickle.load(pkl_file, encoding="latin1"))
                self.radend_dict = manager.dict(
                    pickle.load(pkl_file, encoding="latin1"))
                self.phistart_dict = manager.dict(
                    pickle.load(pkl_file, encoding="latin1"))
                self.phiend_dict = manager.dict(
                    pickle.load(pkl_file, encoding="latin1"))
                self.input_img_num_dict = manager.dict(
                    pickle.load(pkl_file, encoding="latin1"))

                self.klparam_dict = pickle.load(pkl_file, encoding="latin1")

            else:
                self.aligned_images_dict = manager.dict(pickle.load(pkl_file))

                self.klmodes_dict = manager.dict(pickle.load(pkl_file))
                self.evecs_dict = manager.dict(pickle.load(pkl_file))
                self.evals_dict = manager.dict(pickle.load(pkl_file))
                self.ref_psfs_indicies_dict = manager.dict(
                    pickle.load(pkl_file))
                self.section_ind_dict = manager.dict(pickle.load(pkl_file))

                self.radstart_dict = manager.dict(pickle.load(pkl_file))
                self.radend_dict = manager.dict(pickle.load(pkl_file))
                self.phistart_dict = manager.dict(pickle.load(pkl_file))
                self.phiend_dict = manager.dict(pickle.load(pkl_file))
                self.input_img_num_dict = manager.dict(pickle.load(pkl_file))

                self.klparam_dict = pickle.load(pkl_file)

        else:

            if file_extension == ".h5":
                kl_basis_file = _load_dict_from_hdf5(self.basis_filename)

        self.aligned_images_dict = manager.dict(
            kl_basis_file['aligned_images_dict'])

        self.klmodes_dict = manager.dict(kl_basis_file['klmodes_dict'])
        self.evecs_dict = manager.dict(kl_basis_file['evecs_dict'])
        self.evals_dict = manager.dict(kl_basis_file['evals_dict'])
        self.ref_psfs_indicies_dict = manager.dict(
            kl_basis_file['ref_psfs_indicies_dict'])
        self.section_ind_dict = manager.dict(kl_basis_file['section_ind_dict'])

        self.radstart_dict = manager.dict(kl_basis_file['radstart_dict'])
        self.radend_dict = manager.dict(kl_basis_file['radend_dict'])
        self.phistart_dict = manager.dict(kl_basis_file['phistart_dict'])
        self.phiend_dict = manager.dict(kl_basis_file['phiend_dict'])
        self.input_img_num_dict = manager.dict(
            kl_basis_file['input_img_num_dict'])

        self.klparam_dict = kl_basis_file['klparam_dict']

        del kl_basis_file

        # read key name for each section and image
        self.dict_keys = sorted(self.klmodes_dict.keys())

        # load parameters of the correction that fm.klip_dataset produced
        # when we saved the FM basis.

        self.isRDI = (self.klparam_dict['isRDI'] == 1)
        self.IWA = self.klparam_dict['IWA']
        self.OWA = self.klparam_dict['OWA']

        numbasis = self.klparam_dict['numbasis'].astype(int)
        if hasattr(numbasis, "__len__"):
            numbasis = np.array(numbasis)
        else:
            numbasis = np.array([numbasis])

        self.numbasis = numbasis

        self.aligned_center = [
            self.klparam_dict['aligned_center_x'],
            self.klparam_dict['aligned_center_y'],
        ]

        output_imgs_shape = tuple(
            self.klparam_dict['output_imgs_shape'].astype(int))

        self.output_imgs_shape = output_imgs_shape

        # Those are loaded to avoid depending at all on the dataset when we load the KL basis
        self.PAs = self.klparam_dict['PAs']
        self.wvs = self.klparam_dict['wvs']

        self.nwvs = int(self.klparam_dict['nwvs'])  # Get the number of wvls
        self.nfiles = int(
            self.klparam_dict['nfiles'])  # Get the number of wvls

        dim_frame = self.klparam_dict['input_img_shape']
        self.inputs_shape = np.array(
            (self.nfiles * self.nwvs, int(dim_frame[0]), int(dim_frame[1])))

        # After loading it, we stop saving the KL basis to avoid saving it every time
        # we run self.fm_parallelize.
        self.save_basis = False

    def fm_parallelized(self):
        """
        Functions like fm.klip_dataset, but it uses previously measured KL modes,
        section positions, and klip parameter to return the forward modelling.
        Do not save fits.

        Args:
            None

        Returns:
            fmout_np, a numpy array, output of forward modelling
                    * if N_wl = 1, size is [n_KL,x,y]
                    * if N_wl > 1, size is  [n_KL,N_wl,x,y]

        """

        fmout_data, fmout_shape = self.alloc_fmout(self.output_imgs_shape)
        fmout_np = fm._arraytonumpy(fmout_data,
                                    fmout_shape,
                                    dtype=self.data_type)
        # this line is added to be able to use fm._save_rotated_section
        # which uses global var outputs_shape
        fm.outputs_shape = self.output_imgs_shape

        wvs = self.wvs

        if self.isRDI:
            mode = 'RDI'
        else:
            mode = None
            # We are only interested in the RDI mode
            # if not we don't care since it does not have an
            # impact at this point

        for key in self.dict_keys:  # loop pver the sections/images

            img_num = self.input_img_num_dict[key]

            # To have a single identifier for each set of aligned images,
            # we save the wavelenght in nm
            wl_here = wvs[img_num]
            wlstr = 'wl' + str(int(wl_here * 1000)).zfill(4)

            # in load mode, we do not pass aligned_images_dict
            # because it is already in the class to
            # save memory
            self.fm_from_eigen(
                klmodes=self.klmodes_dict[key],
                evals=self.evals_dict[key],
                evecs=self.evecs_dict[key],
                input_img_shape=[self.inputs_shape[1], self.inputs_shape[2]],
                output_img_shape=self.output_imgs_shape,
                input_img_num=img_num,
                ref_psfs_indicies=self.ref_psfs_indicies_dict[key],
                section_ind=self.section_ind_dict[key],
                radstart=self.radstart_dict[key],
                radend=self.radend_dict[key],
                phistart=self.phistart_dict[key],
                phiend=self.phiend_dict[key],
                padding=0.0,
                IOWA=(self.IWA, self.OWA),
                ref_center=self.aligned_center,
                parang=self.PAs[img_num],
                numbasis=self.numbasis,
                fmout=fmout_np,
                mode=mode)

        # put any finishing touches on the FM Output
        fmout_np = fm._arraytonumpy(fmout_data,
                                    fmout_shape,
                                    dtype=self.data_type)
        fmout_np = self.cleanup_fmout(fmout_np)

        # Check if we have a disk model at multiple wavelengths.
        # If true then it's a non- collapsed spec mode disk and we need to reorganise
        # fmout_return. We use the same mean so that it corresponds to
        # klip image-speccube.fits produced by.fm.klip_dataset
        if np.size(np.shape(self.model_disk)) > 2:

            n_wv_per_file = self.nwvs  # Number of WL per file.

            # Collapse across all files, keeping the wavelengths intact.
            fmout_return = np.zeros([
                np.size(self.numbasis),
                n_wv_per_file,
                self.inputs_shape[1],
                self.inputs_shape[2],
            ])
            for i in np.arange(n_wv_per_file):
                fmout_return[:, i, :, :] = np.nansum(
                    fmout_np[:, i::n_wv_per_file, :, :], axis=1) / float(
                        self.nfiles)

        else:
            # If false then this is a collapsed-spec mode or pol mode: collapsed
            # across all files
            fmout_return = np.nanmean(fmout_np, axis=1)

        return fmout_return


##############################################################################
###### 4 routines to save and load h5 in dictionnaries
##############################################################################


def _save_dict_to_hdf5(dic, filename):
    """
    Saving a nested dictionnary into a h5 file

    Args:
        dic: the dictionnary to file
        filename: the filename of the h5 where it will be saved

    Returns:
        None

    """
    with h5py.File(filename, "w") as h5file:
        _recursively_save_dict_contents_to_group(h5file, '/', dic)


def _load_dict_from_hdf5(filename):
    """
    Load a dictionnary from a h5 file

    Args:
        filename: the filename of the h5

    Returns:
        the dictionnary exctracted

    """

    with h5py.File(filename, "r") as h5file:
        return _recursively_load_dict_contents_from_group(h5file, '/')


def _recursively_save_dict_contents_to_group(h5file, path, dic):
    """
    Recursively explore the dictionnary for saving it

    Args:
        h5file: the file in which we save, opened with h5py.File
        path: the separator to aggregate the keys. Should not be set to a value that
            is likely to be in the dictionnary keys already
        dic: the dictionnary to deconstruct

    Returns
        None

    """
    for key, item in dic.items():
        if isinstance(item, (np.ndarray, np.int64, np.float64, str, bytes)):
            h5file[path + key] = item
        elif isinstance(item, dict):
            _recursively_save_dict_contents_to_group(h5file, path + key + '/',
                                                     item)
        else:
            raise ValueError("Cannot save {0} type in h5 (key = {1})".format(
                type(item), path + key))


def _recursively_load_dict_contents_from_group(h5file, path):
    """
    Recursively explore the dictionnary for loading it

    Args:
        h5file: the file from which we load, opened with h5py.File
        path: the separator to aggregate the keys. Should be the same one used in
        _recursively_save_dict_contents_to_group
    Returns
        the rebuilt dictionnary
    """

    ans = {}
    for key, item in h5file[path].items():
        if isinstance(item, h5py._hl.dataset.Dataset):
            ans[key] = item[()]
        elif isinstance(item, h5py._hl.group.Group):
            ans[key] = _recursively_load_dict_contents_from_group(
                h5file, path + key + '/')
    return ans
